"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import copy
import os

import jinja2.exceptions

import salt.modules.rh_ip as rh_ip
import salt.modules.systemd_service as service_mod
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, create_autospec, patch
from tests.support.unit import TestCase


class RhipTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.rh_ip
    """

    def setup_loader_modules(self):
        return {rh_ip: {"__grains__": {"os": "CentOS"}}}

    def test_error_message_iface_should_process_non_str_expected(self):
        values = [1, True, False, "no-kaboom"]
        iface = "ethtest"
        option = "test"
        msg = rh_ip._error_msg_iface(iface, option, values)
        self.assertTrue(msg.endswith("[1|True|False|no-kaboom]"), msg)

    def test_error_message_network_should_process_non_str_expected(self):
        values = [1, True, False, "no-kaboom"]
        msg = rh_ip._error_msg_network("fnord", values)
        self.assertTrue(msg.endswith("[1|True|False|no-kaboom]"), msg)

    def test_build_interface(self):
        """
        Test to build an interface script for a network interface.
        """
        with patch.dict(rh_ip.__grains__, {"os": "Fedora", "osmajorrelease": 26}):
            with patch.object(rh_ip, "_raise_error_iface", return_value=None):
                self.assertRaises(
                    AttributeError, rh_ip.build_interface, "iface", "slave", True
                )

                with patch.dict(
                    rh_ip.__salt__, {"network.interfaces": lambda: {"eth": True}}
                ):
                    self.assertRaises(
                        AttributeError,
                        rh_ip.build_interface,
                        "iface",
                        "eth",
                        True,
                        netmask="255.255.255.255",
                        prefix=32,
                        test=True,
                    )
                    self.assertRaises(
                        AttributeError,
                        rh_ip.build_interface,
                        "iface",
                        "eth",
                        True,
                        ipaddrs=["A"],
                        test=True,
                    )
                    self.assertRaises(
                        AttributeError,
                        rh_ip.build_interface,
                        "iface",
                        "eth",
                        True,
                        ipv6addrs=["A"],
                        test=True,
                    )

        for osrelease in range(7, 8):
            with patch.dict(
                rh_ip.__grains__,
                {"os": "RedHat", "osrelease": str(osrelease)},
            ):
                with patch.object(rh_ip, "_raise_error_iface", return_value=None):
                    with patch.object(rh_ip, "_parse_settings_bond", MagicMock()):
                        mock = jinja2.exceptions.TemplateNotFound("foo")
                        with patch.object(
                            jinja2.Environment,
                            "get_template",
                            MagicMock(side_effect=mock),
                        ):
                            self.assertEqual(
                                rh_ip.build_interface("iface", "vlan", True), ""
                            )

                        with patch.object(rh_ip, "_read_temp", return_value="A"):
                            with patch.object(
                                jinja2.Environment, "get_template", MagicMock()
                            ):
                                self.assertEqual(
                                    rh_ip.build_interface(
                                        "iface", "vlan", True, test="A"
                                    ),
                                    "A",
                                )

                                with patch.object(
                                    rh_ip, "_write_file_iface", return_value=None
                                ):
                                    with patch.object(
                                        os.path, "join", return_value="A"
                                    ):
                                        with patch.object(
                                            rh_ip, "_read_file", return_value="A"
                                        ):
                                            self.assertEqual(
                                                rh_ip.build_interface(
                                                    "iface", "vlan", True
                                                ),
                                                "A",
                                            )
                                            if osrelease > 6:
                                                with patch.dict(
                                                    rh_ip.__salt__,
                                                    {
                                                        "network.interfaces": lambda: {
                                                            "eth": True
                                                        }
                                                    },
                                                ):
                                                    self.assertEqual(
                                                        rh_ip.build_interface(
                                                            "iface",
                                                            "eth",
                                                            True,
                                                            ipaddrs=["127.0.0.1/8"],
                                                        ),
                                                        "A",
                                                    )
                                                    self.assertEqual(
                                                        rh_ip.build_interface(
                                                            "iface",
                                                            "eth",
                                                            True,
                                                            ipv6addrs=["fc00::1/128"],
                                                        ),
                                                        "A",
                                                    )

    def test_build_routes(self):
        """
        Test to build a route script for a network interface.
        """
        with patch.dict(rh_ip.__grains__, {"osrelease": "5.0"}):
            with patch.object(rh_ip, "_parse_routes", MagicMock()):
                mock = jinja2.exceptions.TemplateNotFound("foo")
                with patch.object(
                    jinja2.Environment, "get_template", MagicMock(side_effect=mock)
                ):
                    self.assertEqual(rh_ip.build_routes("iface"), "")

                with patch.object(jinja2.Environment, "get_template", MagicMock()):
                    with patch.object(rh_ip, "_read_temp", return_value=["A"]):
                        self.assertEqual(rh_ip.build_routes("i", test="t"), ["A", "A"])

                    with patch.object(rh_ip, "_read_file", return_value=["A"]):
                        with patch.object(os.path, "join", return_value="A"):
                            with patch.object(
                                rh_ip, "_write_file_iface", return_value=None
                            ):
                                self.assertEqual(
                                    rh_ip.build_routes("i", test=None), ["A", "A"]
                                )

    def test_down(self):
        """
        Test to shutdown a network interface
        """
        with patch.dict(rh_ip.__salt__, {"cmd.run": MagicMock(return_value="A")}):
            self.assertEqual(rh_ip.down("iface", "iface_type"), "A")

        self.assertEqual(rh_ip.down("iface", "slave"), None)

    def test_get_interface(self):
        """
        Test to return the contents of an interface script
        """
        with patch.object(os.path, "join", return_value="A"):
            with patch.object(rh_ip, "_read_file", return_value="A"):
                self.assertEqual(rh_ip.get_interface("iface"), "A")

    def test__parse_settings_eth_hwaddr_and_macaddr(self):
        """
        Test that an AttributeError is thrown when hwaddr and macaddr are
        passed together. They cannot be used together
        """
        opts = {"hwaddr": 1, "macaddr": 2}

        self.assertRaises(
            AttributeError,
            rh_ip._parse_settings_eth,
            opts=opts,
            iface_type="eth",
            enabled=True,
            iface="eth0",
        )

    def test__parse_settings_eth_hwaddr(self):
        """
        Make sure hwaddr gets added when parsing opts
        """
        opts = {"hwaddr": "AA:BB:CC:11:22:33"}
        with patch.dict(rh_ip.__salt__, {"network.interfaces": MagicMock()}):
            results = rh_ip._parse_settings_eth(
                opts=opts, iface_type="eth", enabled=True, iface="eth0"
            )
        self.assertIn("hwaddr", results)
        self.assertEqual(results["hwaddr"], opts["hwaddr"])

    def test__parse_settings_eth_macaddr(self):
        """
        Make sure macaddr gets added when parsing opts
        """
        opts = {"macaddr": "AA:BB:CC:11:22:33"}
        with patch.dict(rh_ip.__salt__, {"network.interfaces": MagicMock()}):
            results = rh_ip._parse_settings_eth(
                opts=opts, iface_type="eth", enabled=True, iface="eth0"
            )
        self.assertIn("macaddr", results)
        self.assertEqual(results["macaddr"], opts["macaddr"])

    def test__parse_settings_eth_ethtool_channels(self):
        """
        Make sure channels gets added when parsing opts
        """
        opts = {"channels": {"rx": 4, "tx": 4, "combined": 4, "other": 4}}
        with patch.dict(rh_ip.__grains__, {"num_cpus": 4}), patch.dict(
            rh_ip.__salt__, {"network.interfaces": MagicMock()}
        ):
            results = rh_ip._parse_settings_eth(
                opts=opts, iface_type="eth", enabled=True, iface="eth0"
            )
        self.assertIn("ethtool", results)
        self.assertEqual(results["ethtool"], "-L eth0 rx 4 tx 4 other 4 combined 4")

    def test_up(self):
        """
        Test to start up a network interface
        """
        with patch.dict(rh_ip.__salt__, {"cmd.run": MagicMock(return_value="A")}):
            self.assertEqual(rh_ip.up("iface", "iface_type"), "A")

        self.assertEqual(rh_ip.up("iface", "slave"), None)

    def test_get_routes(self):
        """
        Test to return the contents of the interface routes script.
        """
        with patch.object(os.path, "join", return_value="A"):
            with patch.object(rh_ip, "_read_file", return_value=["A"]):
                self.assertEqual(rh_ip.get_routes("iface"), ["A", "A"])

    def test_get_network_settings(self):
        """
        Test to return the contents of the global network script.
        """
        with patch.object(rh_ip, "_read_file", return_value="A"):
            self.assertEqual(rh_ip.get_network_settings(), "A")

    def test_apply_network_settings(self):
        """
        Test to apply global network configuration.
        """
        # This should be pytest.mark.parametrize, when this gets ported to
        # pytest approach. This is just following previous patterns here.
        # Edge cases are 7 & 8
        mock_service = create_autospec(service_mod.restart, return_value=True)
        for majorrelease, expected_service_name in (
            (3, "network"),
            (7, "network"),
            (8, "NetworkManager"),
            (42, "NetworkManager"),
        ):
            with patch.dict(
                rh_ip.__salt__, {"service.restart": mock_service}
            ), patch.dict(
                rh_ip.__grains__,
                {"osmajorrelease": majorrelease},
            ):
                self.assertTrue(rh_ip.apply_network_settings())
                mock_service.assert_called_with(expected_service_name)

    def test_build_network_settings(self):
        """
        Test to build the global network script.
        """
        with patch.object(rh_ip, "_parse_rh_config", MagicMock()):
            with patch.object(rh_ip, "_parse_network_settings", MagicMock()):

                mock = jinja2.exceptions.TemplateNotFound("foo")
                with patch.object(
                    jinja2.Environment, "get_template", MagicMock(side_effect=mock)
                ):
                    self.assertEqual(rh_ip.build_network_settings(), "")

                with patch.object(jinja2.Environment, "get_template", MagicMock()):
                    with patch.object(rh_ip, "_read_temp", return_value="A"):
                        self.assertEqual(rh_ip.build_network_settings(test="t"), "A")

                        with patch.object(
                            rh_ip, "_write_file_network", return_value=None
                        ):
                            with patch.object(rh_ip, "_read_file", return_value="A"):
                                self.assertEqual(
                                    rh_ip.build_network_settings(test=None), "A"
                                )

    def test_build_interface_teamport(self):
        """
        Test that teamport interfaces are properly built
        """
        ifaces = MagicMock(return_value={"eth1": {"hwaddr": "02:42:ac:11:00:02"}})
        dunder_salt = {"network.interfaces": ifaces}
        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__, {"osmajorrelease": version, "osrelease": str(version)}
            ), patch.dict(rh_ip.__salt__, dunder_salt):
                ret = sorted(
                    rh_ip.build_interface(
                        "eth1",
                        "teamport",
                        enabled=True,
                        test=True,
                        team_port_config={"prio": 100},
                        team_master="team0",
                    )
                )

            expected = [
                'DEVICE="eth1"',
                'DEVICETYPE="TeamPort"',
                'HWADDR="02:42:ac:11:00:02"',
                'NM_CONTROLLED="no"',
                'ONBOOT="yes"',
                'TEAM_MASTER="team0"',
                "TEAM_PORT_CONFIG='{\"prio\": 100}'",
                'USERCTL="no"',
            ]
            assert ret == expected, ret

    def test_build_interface_team(self):
        """
        Test that team interfaces are properly built
        """
        dunder_salt = {"pkg.version": MagicMock(return_value="1.29-1.el7")}
        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__, {"osmajorrelease": version, "osrelease": str(version)}
            ), patch.dict(rh_ip.__salt__, dunder_salt):
                ret = sorted(
                    rh_ip.build_interface(
                        "team0",
                        "team",
                        enabled=True,
                        test=True,
                        ipaddr="1.2.3.4",
                        team_config={"foo": "bar"},
                    )
                )

            expected = [
                'DEVICE="team0"',
                'DEVICETYPE="Team"',
                'IPADDR="1.2.3.4"',
                'NM_CONTROLLED="no"',
                'ONBOOT="yes"',
                'TEAM_CONFIG=\'{"foo": "bar"}\'',
                'USERCTL="no"',
            ]
            assert ret == expected

    @staticmethod
    def _check_common_opts_bond(lines):
        """
        Reduce code duplication by making sure that the expected options are
        present in the config file. Note that this assumes that duplex="full"
        was passed in the kwargs. If it wasn't, then there would be no
        ETHTOOL_OPTS line.
        """
        assert 'DEVICE="bond0"' in lines
        assert 'ETHTOOL_OPTS="duplex full"' in lines
        assert 'NM_CONTROLLED="no"' in lines
        assert 'ONBOOT="yes"' in lines
        assert 'TYPE="Bond"' in lines
        assert 'USERCTL="no"' in lines

    def _validate_miimon_downdelay(self, kwargs):
        """
        Validate that downdelay that is not a multiple of miimon raises an error
        """
        # Make copy of kwargs so we don't modify what was passed in
        kwargs = copy.copy(kwargs)

        # Remove miimon and downdelay (if present) to test invalid input
        for key in ("miimon", "downdelay"):
            kwargs.pop(key, None)

        kwargs["miimon"] = 100
        kwargs["downdelay"] = 201
        try:
            rh_ip.build_interface(
                "bond0",
                "bond",
                enabled=True,
                **kwargs,
            )
        except AttributeError as exc:
            assert "multiple of miimon" in str(exc)
        else:
            raise Exception("AttributeError was not raised")

    def _validate_miimon_conf(self, kwargs, required=True):
        """
        Validate miimon configuration
        """
        # Make copy of kwargs so we don't modify what was passed in
        kwargs = copy.copy(kwargs)

        # Remove miimon and downdelay (if present) to test invalid input
        for key in ("miimon", "downdelay"):
            kwargs.pop(key, None)

        if required:
            # Leaving out miimon should raise an error
            try:
                rh_ip.build_interface(
                    "bond0",
                    "bond",
                    enabled=True,
                    **kwargs,
                )
            except AttributeError as exc:
                assert "miimon" in str(exc)
            else:
                raise Exception("AttributeError was not raised")

        self._validate_miimon_downdelay(kwargs)

    def _get_bonding_opts(self, kwargs):
        results = rh_ip.build_interface(
            "bond0",
            "bond",
            enabled=True,
            **kwargs,
        )
        self._check_common_opts_bond(results)

        for line in results:
            if line.startswith("BONDING_OPTS="):
                return sorted(line.split("=", 1)[-1].strip('"').split())
        raise Exception("BONDING_OPTS not found")

    def _test_mode_0_or_2(self, mode_num=0):
        """
        Modes 0 and 2 share the majority of code, with mode 2 being a superset
        of mode 0. This function will do the proper asserts for the common code
        in these two modes.
        """
        kwargs = {
            "test": True,
            "duplex": "full",
            "slaves": "eth1 eth2",
        }

        if mode_num == 0:
            modes = ("balance-rr", mode_num, str(mode_num))
        else:
            modes = ("balance-xor", mode_num, str(mode_num))

        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__, {"osmajorrelease": version, "osrelease": str(version)}
            ):
                for mode in modes:
                    kwargs["mode"] = mode
                    # Remove all miimon/arp settings to test invalid config
                    for key in (
                        "miimon",
                        "downdelay",
                        "arp_interval",
                        "arp_ip_targets",
                    ):
                        kwargs.pop(key, None)

                    # Check that invalid downdelay is handled correctly
                    self._validate_miimon_downdelay(kwargs)

                    # Leaving out miimon and arp_interval should raise an error
                    try:
                        bonding_opts = self._get_bonding_opts(kwargs)
                    except AttributeError as exc:
                        assert "miimon or arp_interval" in str(exc)
                    else:
                        raise Exception("AttributeError was not raised")

                    kwargs["miimon"] = 100
                    kwargs["downdelay"] = 200
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "downdelay=200",
                        "miimon=100",
                        "mode={}".format(mode_num),
                        "use_carrier=0",
                    ]
                    assert bonding_opts == expected, bonding_opts

                    # Add arp settings, and test again
                    kwargs["arp_interval"] = 300
                    kwargs["arp_ip_target"] = ["1.2.3.4", "5.6.7.8"]
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "arp_interval=300",
                        "arp_ip_target=1.2.3.4,5.6.7.8",
                        "downdelay=200",
                        "miimon=100",
                        "mode={}".format(mode_num),
                        "use_carrier=0",
                    ]
                    assert bonding_opts == expected, bonding_opts

                    # Remove miimon and downdelay and test again
                    del kwargs["miimon"]
                    del kwargs["downdelay"]
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "arp_interval=300",
                        "arp_ip_target=1.2.3.4,5.6.7.8",
                        "mode={}".format(mode_num),
                    ]
                    assert bonding_opts == expected, bonding_opts

    def test_build_interface_bond_mode_0(self):
        """
        Test that mode 0 bond interfaces are properly built
        """
        self._test_mode_0_or_2(0)

    def test_build_interface_bond_mode_1(self):
        """
        Test that mode 1 bond interfaces are properly built
        """
        kwargs = {
            "test": True,
            "mode": "active-backup",
            "duplex": "full",
            "slaves": "eth1 eth2",
            "miimon": 100,
            "downdelay": 200,
        }

        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__, {"osmajorrelease": version, "osrelease": str(version)}
            ):
                for mode in ("active-backup", 1, "1"):
                    kwargs.pop("primary", None)
                    kwargs["mode"] = mode
                    self._validate_miimon_conf(kwargs)
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "downdelay=200",
                        "miimon=100",
                        "mode=1",
                        "use_carrier=0",
                    ]
                    assert bonding_opts == expected, bonding_opts

                    # Add a "primary" option and confirm that it shows up in
                    # the bonding opts.
                    kwargs["primary"] = "foo"
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "downdelay=200",
                        "miimon=100",
                        "mode=1",
                        "primary=foo",
                        "use_carrier=0",
                    ]
                    assert bonding_opts == expected, bonding_opts

    def test_build_interface_bond_mode_2(self):
        """
        Test that mode 2 bond interfaces are properly built
        """
        self._test_mode_0_or_2(2)

        kwargs = {
            "test": True,
            "duplex": "full",
            "slaves": "eth1 eth2",
            "miimon": 100,
            "downdelay": 200,
        }
        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__, {"osmajorrelease": version, "osrelease": str(version)}
            ):
                for mode in ("balance-xor", 2, "2"):
                    # Using an invalid hashing algorithm should cause an error
                    # to be raised.
                    kwargs["mode"] = mode
                    kwargs["hashing-algorithm"] = "layer42"
                    try:
                        bonding_opts = self._get_bonding_opts(kwargs)
                    except AttributeError as exc:
                        assert "hashing-algorithm" in str(exc)
                    else:
                        raise Exception("AttributeError was not raised")

                    # Correct the hashing algorithm and re-run
                    kwargs["hashing-algorithm"] = "layer2"
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "downdelay=200",
                        "miimon=100",
                        "mode=2",
                        "use_carrier=0",
                        "xmit_hash_policy=layer2",
                    ]
                    assert bonding_opts == expected, bonding_opts

    def test_build_interface_bond_mode_3(self):
        """
        Test that mode 3 bond interfaces are properly built
        """
        kwargs = {
            "test": True,
            "duplex": "full",
            "slaves": "eth1 eth2",
            "miimon": 100,
            "downdelay": 200,
        }

        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__, {"osmajorrelease": version, "osrelease": str(version)}
            ):
                for mode in ("broadcast", 3, "3"):
                    kwargs["mode"] = mode
                    self._validate_miimon_conf(kwargs)
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "downdelay=200",
                        "miimon=100",
                        "mode=3",
                        "use_carrier=0",
                    ]
                    assert bonding_opts == expected, bonding_opts

    def test_build_interface_bond_mode_4_xmit(self):
        """
        Test that mode 4 bond interfaces are properly built
        """
        kwargs = {
            "test": True,
            "duplex": "full",
            "slaves": "eth1 eth2",
            "miimon": 100,
            "downdelay": 200,
        }
        valid_lacp_rate = ("fast", "slow", "1", "0")
        valid_ad_select = ("0",)

        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__,
                {
                    "osmajorrelease": version,
                    "osrelease": str(version),
                    "os_family": "RedHat",
                },
            ):
                for mode in ("802.3ad", 4, "4"):
                    kwargs["mode"] = mode
                    self._validate_miimon_conf(kwargs)

                    for version in range(7, 8):
                        with patch.dict(rh_ip.__grains__, {"osmajorrelease": version}):
                            # Using an invalid hashing algorithm should cause an error
                            # to be raised.
                            kwargs["hashing-algorithm"] = "layer42"
                            try:
                                bonding_opts = self._get_bonding_opts(kwargs)
                            except AttributeError as exc:
                                assert "hashing-algorithm" in str(exc)
                            else:
                                raise Exception("AttributeError was not raised")

                        hash_alg = "vlan+srcmac"
                        if version == 7:
                            # Using an invalid hashing algorithm should cause an error
                            # to be raised.
                            kwargs["hashing-algorithm"] = hash_alg
                            try:
                                bonding_opts = self._get_bonding_opts(kwargs)
                            except AttributeError as exc:
                                assert "hashing-algorithm" in str(exc)
                            else:
                                raise Exception("AttributeError was not raised")
                        else:
                            # Correct the hashing algorithm and re-run
                            kwargs["hashing-algorithm"] = hash_alg
                            bonding_opts = self._get_bonding_opts(kwargs)
                            expected = [
                                "ad_select=0",
                                "downdelay=200",
                                "lacp_rate=0",
                                "miimon=100",
                                "mode=4",
                                "use_carrier=0",
                                "xmit_hash_policy={}".format(hash_alg),
                            ]
                            assert bonding_opts == expected, bonding_opts

                        for hash_alg in [
                            "layer2",
                            "layer2+3",
                            "layer3+4",
                            "encap2+3",
                            "encap3+4",
                        ]:
                            # Correct the hashing algorithm and re-run
                            kwargs["hashing-algorithm"] = hash_alg
                            bonding_opts = self._get_bonding_opts(kwargs)
                            expected = [
                                "ad_select=0",
                                "downdelay=200",
                                "lacp_rate=0",
                                "miimon=100",
                                "mode=4",
                                "use_carrier=0",
                                "xmit_hash_policy={}".format(hash_alg),
                            ]
                            assert bonding_opts == expected, bonding_opts

    def test_build_interface_bond_mode_4_lacp(self):
        """
        Test that mode 4 bond interfaces are properly built
        """
        kwargs = {
            "test": True,
            "duplex": "full",
            "slaves": "eth1 eth2",
            "miimon": 100,
            "downdelay": 200,
        }
        valid_lacp_rate = ("fast", "slow", "1", "0")
        valid_ad_select = ("0",)

        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__, {"osmajorrelease": version, "osrelease": str(version)}
            ):
                for mode in ("802.3ad", 4, "4"):
                    kwargs["mode"] = mode
                    self._validate_miimon_conf(kwargs)

                    for lacp_rate in valid_lacp_rate + ("2", "speedy"):
                        for ad_select in valid_ad_select + ("foo",):
                            kwargs["lacp_rate"] = lacp_rate
                            kwargs["ad_select"] = ad_select
                            try:
                                bonding_opts = self._get_bonding_opts(kwargs)
                            except AttributeError as exc:
                                error = str(exc)
                                # Re-raise the exception only if it was
                                # unexpected. It should not be expected when
                                # the lacp_rate or ad_select is valid.
                                if "lacp_rate" in error:
                                    if lacp_rate in valid_lacp_rate:
                                        raise
                                elif "ad_select" in error:
                                    if ad_select in valid_ad_select:
                                        raise
                                else:
                                    raise
                            else:
                                expected = [
                                    "ad_select={}".format(ad_select),
                                    "downdelay=200",
                                    "lacp_rate={}".format(
                                        "1"
                                        if lacp_rate == "fast"
                                        else "0"
                                        if lacp_rate == "slow"
                                        else lacp_rate
                                    ),
                                    "miimon=100",
                                    "mode=4",
                                    "use_carrier=0",
                                ]
                                assert bonding_opts == expected, bonding_opts

    def test_build_interface_bond_mode_5(self):
        """
        Test that mode 5 bond interfaces are properly built
        """
        kwargs = {
            "test": True,
            "duplex": "full",
            "slaves": "eth1 eth2",
            "miimon": 100,
            "downdelay": 200,
        }

        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__, {"osmajorrelease": version, "osrelease": str(version)}
            ):
                for mode in ("balance-tlb", 5, "5"):
                    kwargs.pop("primary", None)
                    kwargs["mode"] = mode
                    self._validate_miimon_conf(kwargs)
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "downdelay=200",
                        "miimon=100",
                        "mode=5",
                        "use_carrier=0",
                    ]
                    assert bonding_opts == expected, bonding_opts

                    # Add a "primary" option and confirm that it shows up in
                    # the bonding opts.
                    kwargs["primary"] = "foo"
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "downdelay=200",
                        "miimon=100",
                        "mode=5",
                        "primary=foo",
                        "use_carrier=0",
                    ]
                    assert bonding_opts == expected, bonding_opts

    def test_build_interface_bond_mode_6(self):
        """
        Test that mode 6 bond interfaces are properly built
        """
        kwargs = {
            "test": True,
            "duplex": "full",
            "slaves": "eth1 eth2",
            "miimon": 100,
            "downdelay": 200,
        }

        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__, {"osmajorrelease": version, "osrelease": str(version)}
            ):
                for mode in ("balance-alb", 6, "6"):
                    kwargs.pop("primary", None)
                    kwargs["mode"] = mode
                    self._validate_miimon_conf(kwargs)
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "downdelay=200",
                        "miimon=100",
                        "mode=6",
                        "use_carrier=0",
                    ]
                    assert bonding_opts == expected, bonding_opts

                    # Add a "primary" option and confirm that it shows up in
                    # the bonding opts.
                    kwargs["primary"] = "foo"
                    bonding_opts = self._get_bonding_opts(kwargs)
                    expected = [
                        "downdelay=200",
                        "miimon=100",
                        "mode=6",
                        "primary=foo",
                        "use_carrier=0",
                    ]
                    assert bonding_opts == expected, bonding_opts

    def test_build_interface_bond_slave(self):
        """
        Test that bond slave interfaces are properly built
        """
        for version in range(7, 8):
            with patch.dict(
                rh_ip.__grains__, {"osmajorrelease": version, "osrelease": str(version)}
            ):
                results = sorted(
                    rh_ip.build_interface(
                        "eth1",
                        "slave",
                        enabled=True,
                        test=True,
                        master="bond0",
                    )
                )
                expected = [
                    'BOOTPROTO="none"',
                    'DEVICE="eth1"',
                    'MASTER="bond0"',
                    'NM_CONTROLLED="no"',
                    'ONBOOT="yes"',
                    'SLAVE="yes"',
                    'USERCTL="no"',
                ]
                assert results == expected, results



"""
Service support for Solaris 10 and 11, should work with other systems
that use SMF also. (e.g. SmartOS)

.. important::
    If you feel that Salt should be using this module to manage services on a
    minion, and it is using a different module (or gives an error similar to
    *'service.start' is not available*), see :ref:`here
    <module-provider-override>`.
"""


import fnmatch
import re

__func_alias__ = {"reload_": "reload"}

# Define the module's virtual name
__virtualname__ = "service"


def __virtual__():
    """
    Only work on systems which default to SMF
    """
    if "Solaris" in __grains__["os_family"]:
        # Don't let this work on Solaris 9 since SMF doesn't exist on it.
        if __grains__["kernelrelease"] == "5.9":
            return (
                False,
                "The smf execution module failed to load: SMF not available on"
                " Solaris 9.",
            )
        return __virtualname__
    return (
        False,
        "The smf execution module failed to load: only available on Solaris.",
    )


def _get_enabled_disabled(enabled_prop="true"):
    """
    DRY: Get all service FMRIs and their enabled property
    """
    ret = set()
    cmd = '/usr/bin/svcprop -c -p general/enabled "*"'
    lines = __salt__["cmd.run_stdout"](cmd, python_shell=False).splitlines()
    for line in lines:
        comps = line.split()
        if not comps:
            continue
        if comps[2] == enabled_prop:
            ret.add(comps[0].split("/:properties")[0])
    return sorted(ret)


def get_running():
    """
    Return the running services

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_running
    """
    ret = set()
    cmd = "/usr/bin/svcs -H -o FMRI,STATE -s FMRI"
    lines = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    for line in lines:
        comps = line.split()
        if not comps:
            continue
        if "online" in line:
            ret.add(comps[0])
    return sorted(ret)


def get_stopped():
    """
    Return the stopped services

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_stopped
    """
    ret = set()
    cmd = "/usr/bin/svcs -aH -o FMRI,STATE -s FMRI"
    lines = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    for line in lines:
        comps = line.split()
        if not comps:
            continue
        if "online" not in line and "legacy_run" not in line:
            ret.add(comps[0])
    return sorted(ret)


def available(name):
    """
    Returns ``True`` if the specified service is available, otherwise returns
    ``False``.

    We look up the name with the svcs command to get back the FMRI
    This allows users to use simpler service names

    CLI Example:

    .. code-block:: bash

        salt '*' service.available net-snmp
    """
    cmd = "/usr/bin/svcs -H -o FMRI {}".format(name)
    name = __salt__["cmd.run"](cmd, python_shell=False)
    return name in get_all()


def missing(name):
    """
    The inverse of service.available.
    Returns ``True`` if the specified service is not available, otherwise returns
    ``False``.

    CLI Example:

    .. code-block:: bash

        salt '*' service.missing net-snmp
    """
    cmd = "/usr/bin/svcs -H -o FMRI {}".format(name)
    name = __salt__["cmd.run"](cmd, python_shell=False)
    return name not in get_all()


def get_all():
    """
    Return all installed services

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_all
    """
    ret = set()
    cmd = "/usr/bin/svcs -aH -o FMRI,STATE -s FMRI"
    lines = __salt__["cmd.run"](cmd).splitlines()
    for line in lines:
        comps = line.split()
        if not comps:
            continue
        ret.add(comps[0])
    return sorted(ret)


def start(name):
    """
    Start the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.start <service name>
    """
    cmd = "/usr/sbin/svcadm enable -s -t {}".format(name)
    retcode = __salt__["cmd.retcode"](cmd, python_shell=False)
    if not retcode:
        return True
    if retcode == 3:
        # Return code 3 means there was a problem with the service
        # A common case is being in the 'maintenance' state
        # Attempt a clear and try one more time
        clear_cmd = "/usr/sbin/svcadm clear {}".format(name)
        __salt__["cmd.retcode"](clear_cmd, python_shell=False)
        return not __salt__["cmd.retcode"](cmd, python_shell=False)
    return False


def stop(name):
    """
    Stop the specified service

    CLI Example:

    .. code-block:: bash

        salt '*' service.stop <service name>
    """
    cmd = "/usr/sbin/svcadm disable -s -t {}".format(name)
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def restart(name):
    """
    Restart the named service

    CLI Example:

    .. code-block:: bash

        salt '*' service.restart <service name>
    """
    cmd = "/usr/sbin/svcadm restart {}".format(name)
    if not __salt__["cmd.retcode"](cmd, python_shell=False):
        # calling restart doesn't clear maintenance
        # or tell us that the service is in the 'online' state
        return start(name)
    return False


def reload_(name):
    """
    Reload the named service

    CLI Example:

    .. code-block:: bash

        salt '*' service.reload <service name>
    """
    cmd = "/usr/sbin/svcadm refresh {}".format(name)
    if not __salt__["cmd.retcode"](cmd, python_shell=False):
        # calling reload doesn't clear maintenance
        # or tell us that the service is in the 'online' state
        return start(name)
    return False


def status(name, sig=None):
    """
    Return the status for a service.
    If the name contains globbing, a dict mapping service name to True/False
    values is returned.

    .. versionchanged:: 2018.3.0
        The service name can now be a glob (e.g. ``salt*``)

    Args:
        name (str): The name of the service to check
        sig (str): Not implemented

    Returns:
        bool: True if running, False otherwise
        dict: Maps service name to True if running, False otherwise

    CLI Example:

    .. code-block:: bash

        salt '*' service.status <service name>
    """
    contains_globbing = bool(re.search(r"\*|\?|\[.+\]", name))
    if contains_globbing:
        services = fnmatch.filter(get_all(), name)
    else:
        services = [name]
    results = {}
    for service in services:
        cmd = "/usr/bin/svcs -H -o STATE {}".format(service)
        line = __salt__["cmd.run"](cmd, python_shell=False)
        results[service] = line == "online"
    if contains_globbing:
        return results
    return results[name]


def enable(name, **kwargs):
    """
    Enable the named service to start at boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.enable <service name>
    """
    cmd = "/usr/sbin/svcadm enable {}".format(name)
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def disable(name, **kwargs):
    """
    Disable the named service to start at boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.disable <service name>
    """
    cmd = "/usr/sbin/svcadm disable {}".format(name)
    return not __salt__["cmd.retcode"](cmd, python_shell=False)


def enabled(name, **kwargs):
    """
    Check to see if the named service is enabled to start on boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.enabled <service name>
    """
    # The property that reveals whether a service is enabled
    # can only be queried using the full FMRI
    # We extract the FMRI and then do the query
    fmri_cmd = "/usr/bin/svcs -H -o FMRI {}".format(name)
    fmri = __salt__["cmd.run"](fmri_cmd, python_shell=False)
    cmd = "/usr/sbin/svccfg -s {} listprop general/enabled".format(fmri)
    comps = __salt__["cmd.run"](cmd, python_shell=False).split()
    if comps[2] == "true":
        return True
    else:
        return False


def disabled(name):
    """
    Check to see if the named service is disabled to start on boot

    CLI Example:

    .. code-block:: bash

        salt '*' service.disabled <service name>
    """
    return not enabled(name)


def get_enabled():
    """
    Return the enabled services

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_enabled
    """
    # Note that this returns the full FMRI
    return _get_enabled_disabled("true")


def get_disabled():
    """
    Return the disabled services

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_disabled
    """
    # Note that this returns the full FMRI
    return _get_enabled_disabled("false")



import functools
import os
import pathlib
import subprocess
import tempfile

import pytest

import salt.utils.user


def _check_skip(grains):
    if grains["os"] == "MacOS":
        return True
    return False


pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_on_windows,
    pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip),
]


@pytest.fixture(scope="module")
def account_1():
    with pytest.helpers.create_account(create_group=True) as _account:
        yield _account


@pytest.fixture(scope="module")
def account_2(account_1):
    with pytest.helpers.create_account(group_name=account_1.group.name) as _account:
        yield _account


def test_chugid(account_1):

    # Since we're changing accounts to touch the file, the parent directory must be user and group writable
    with tempfile.TemporaryDirectory() as tmp_path:
        tmp_path = pathlib.Path(tmp_path)
        tmp_path.chmod(0o770)

        testfile = tmp_path / "testfile"

        # We should fail because the parent directory group owner is not the account running the test
        ret = subprocess.run(
            ["touch", str(testfile)],
            preexec_fn=functools.partial(
                salt.utils.user.chugid_and_umask,
                runas=account_1.username,
                umask=None,
                group=None,
            ),
            check=False,
        )
        assert ret.returncode != 0

        # However if we change the group ownership to one of the account's groups, it should succeed
        os.chown(str(tmp_path), 0, account_1.group.info.gid)

        ret = subprocess.run(
            ["touch", str(testfile)],
            preexec_fn=functools.partial(
                salt.utils.user.chugid_and_umask,
                runas=account_1.username,
                umask=None,
                group=None,
            ),
            check=False,
        )
        assert ret.returncode == 0
        assert testfile.exists()
        testfile_stat = testfile.stat()
        assert testfile_stat.st_uid == account_1.info.uid
        assert testfile_stat.st_gid == account_1.info.gid


def test_chugid_and_group(account_1, account_2, tmp_path):

    # Since we're changing accounts to touch the file, the parent directory must be world-writable
    with tempfile.TemporaryDirectory() as tmp_path:
        tmp_path = pathlib.Path(tmp_path)

        tmp_path.chmod(0o770)

        testfile = tmp_path / "testfile"

        # We should fail because the parent directory group owner is not the account running the test
        ret = subprocess.run(
            ["touch", str(testfile)],
            preexec_fn=functools.partial(
                salt.utils.user.chugid_and_umask,
                runas=account_2.username,
                umask=None,
                group=account_1.group.name,
            ),
            check=False,
        )
        assert ret.returncode != 0

        # However if we change the group ownership to one of the account's groups, it should succeed
        os.chown(str(tmp_path), 0, account_1.group.info.gid)

        ret = subprocess.run(
            ["touch", str(testfile)],
            preexec_fn=functools.partial(
                salt.utils.user.chugid_and_umask,
                runas=account_2.username,
                umask=None,
                group=account_1.group.name,
            ),
            check=False,
        )
        assert ret.returncode == 0
        assert testfile.exists()
        testfile_stat = testfile.stat()
        assert testfile_stat.st_uid == account_2.info.uid
        assert testfile_stat.st_gid == account_1.group.info.gid



"""
Utilities for managing kickstart

.. versionadded:: 2015.8.0
"""

import argparse
import shlex

import salt.utils.files
import salt.utils.yaml


def clean_args(args):
    """
    Cleans up the args that weren't passed in
    """
    for arg in list(args):
        if not args[arg]:
            del args[arg]
    return args


def parse_auth(rule):
    """
    Parses the auth/authconfig line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    noargs = (
        "back",
        "test",
        "nostart",
        "kickstart",
        "probe",
        "enablecache",
        "disablecache",
        "disablenis",
        "enableshadow",
        "disableshadow",
        "enablemd5",
        "disablemd5",
        "enableldap",
        "enableldapauth",
        "enableldaptls",
        "disableldap",
        "disableldapauth",
        "enablekrb5kdcdns",
        "disablekrb5kdcdns",
        "enablekrb5realmdns",
        "disablekrb5realmdns",
        "disablekrb5",
        "disablehe-siod",
        "enablesmbauth",
        "disablesmbauth",
        "enablewinbind",
        "enablewinbindauth",
        "disablewinbind",
        "disablewinbindauth",
        "enablewinbindusedefaultdomain",
        "disablewinbindusedefaultdomain",
        "enablewins",
        "disablewins",
    )
    for arg in noargs:
        parser.add_argument("--{}".format(arg), dest=arg, action="store_true")

    parser.add_argument("--enablenis", dest="enablenis", action="store")
    parser.add_argument("--hesiodrhs", dest="hesiodrhs", action="store")
    parser.add_argument("--krb5adminserver", dest="krb5adminserver", action="append")
    parser.add_argument("--krb5kdc", dest="krb5kdc", action="append")
    parser.add_argument("--ldapbasedn", dest="ldapbasedn", action="store")
    parser.add_argument("--ldapserver", dest="ldapserver", action="append")
    parser.add_argument("--nisserver", dest="nisserver", action="append")
    parser.add_argument("--passalgo", dest="passalgo", action="store")
    parser.add_argument("--smbidmapgid", dest="smbidmapgid", action="store")
    parser.add_argument("--smbidmapuid", dest="smbidmapuid", action="store")
    parser.add_argument("--smbrealm", dest="smbrealm", action="store")
    parser.add_argument(
        "--smbsecurity",
        dest="smbsecurity",
        action="store",
        choices=["user", "server", "domain", "dns"],
    )
    parser.add_argument("--smbservers", dest="smbservers", action="store")
    parser.add_argument("--smbworkgroup", dest="smbworkgroup", action="store")
    parser.add_argument("--winbindjoin", dest="winbindjoin", action="store")
    parser.add_argument("--winbindseparator", dest="winbindseparator", action="store")
    parser.add_argument(
        "--winbindtemplatehomedir", dest="winbindtemplatehomedir", action="store"
    )
    parser.add_argument(
        "--winbindtemplateprimarygroup",
        dest="winbindtemplateprimarygroup",
        action="store",
    )
    parser.add_argument(
        "--winbindtemplateshell", dest="winbindtemplateshell", action="store"
    )

    parser.add_argument("--enablekrb5", dest="enablekrb5", action="store_true")
    if "--enablekrb5" in rules:
        parser.add_argument(
            "--krb5realm", dest="krb5realm", action="store", required=True
        )
    parser.add_argument("--enablehesiod", dest="enablehesiod", action="store_true")
    if "--enablehesiod" in rules:
        parser.add_argument(
            "--hesiodlhs", dest="hesiodlhs", action="store", required=True
        )

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_autopart(rule):
    """
    Parse the autopart line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--type", dest="type", action="store")
    parser.add_argument("--encrypted", dest="encrypted", action="store_true")
    parser.add_argument("--passphrase", dest="passphrase", action="store")
    parser.add_argument("--escrowcert", dest="escrowcert", action="store")
    parser.add_argument("--backuppassphrase", dest="backuppassphrase", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_autostep(rule):
    """
    Parse the autostep line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--autoscreenshot", dest="autoscreenshot", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_bootloader(rule):
    """
    Parse the bootloader line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--append", dest="append", action="store")
    parser.add_argument("--driveorder", dest="driveorder", action="store")
    parser.add_argument("--location", dest="location", action="store")
    parser.add_argument("--password", dest="password", action="store")
    parser.add_argument("--md5pass", dest="md5pass", action="store")
    parser.add_argument("--upgrade", dest="upgrade", action="store_true")
    parser.add_argument("--timeout", dest="timeout", action="store")
    parser.add_argument("--boot-drive", dest="bootdrive", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_btrfs(rule):
    """
    Parse the btrfs line

    TODO: finish up the weird parsing on this one
    http://fedoraproject.org/wiki/Anaconda/Kickstart#btrfs
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--name", dest="name", action="store")
    parser.add_argument("--data", dest="data", action="store")
    parser.add_argument("--metadata", dest="metadata", action="store")
    parser.add_argument("--label", dest="label", action="store")
    parser.add_argument("--noformat", dest="noformat", action="store_true")
    parser.add_argument("--useexisting", dest="useexisting", action="store_true")
    parser.add_argument("--subvol", dest="subvol", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_clearpart(rule):
    """
    Parse the clearpart line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--all", dest="all", action="store_true")
    parser.add_argument("--drives", dest="drives", action="store")
    parser.add_argument("--init_label", dest="init_label", action="store_true")
    parser.add_argument("--linux", dest="linux", action="store_true")
    parser.add_argument("--none", dest="none", action="store_true")
    parser.add_argument("--initlabel", dest="init_label", action="store_true")
    parser.add_argument("--list", dest="list", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_device(rule):
    """
    Parse the device line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    modulename = rules.pop(0)
    parser.add_argument("--opts", dest="opts", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    args["modulename"] = modulename
    parser = None
    return args


def parse_dmraid(rule):
    """
    Parse the dmraid line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--name", dest="name", action="store")
    parser.add_argument("--dev", dest="dev", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_driverdisk(rule):
    """
    Parse the driverdisk line
    """
    if "--" not in rule:
        return {"partition": rule}

    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--source", dest="source", action="store")
    parser.add_argument("--biospart", dest="biospart", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_firewall(rule):
    """
    Parse the firewall line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--enable", "--enabled", dest="enable", action="store_true")
    parser.add_argument("--disable", "--disabled", dest="disable", action="store_true")
    parser.add_argument("--port", dest="port", action="store")
    parser.add_argument("--service", dest="service", action="store")
    parser.add_argument("--ssh", dest="ssh", action="store_true")
    parser.add_argument("--smtp", dest="smtp", action="store_true")
    parser.add_argument("--http", dest="http", action="store_true")
    parser.add_argument("--ftp", dest="ftp", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_firstboot(rule):
    """
    Parse the firstboot line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--enable", "--enabled", dest="enable", action="store_true")
    parser.add_argument("--disable", "--disabled", dest="disable", action="store_true")
    parser.add_argument("--reconfig", dest="reconfig", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_group(rule):
    """
    Parse the group line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--name", dest="name", action="store")
    parser.add_argument("--gid", dest="gid", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_harddrive(rule):
    """
    Parse the harddrive line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--biospart", dest="biospart", action="store")
    parser.add_argument("--partition", dest="partition", action="store")
    parser.add_argument("--dir", dest="dir", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_ignoredisk(rule):
    """
    Parse the ignoredisk line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--drives", dest="drives", action="store")
    parser.add_argument("--only-use", dest="only-use", action="store")
    parser.add_argument("--interactive", dest="interactive", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_iscsi(rule):
    """
    Parse the iscsi line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--ipaddr", dest="ipaddr", action="store")
    parser.add_argument("--port", dest="port", action="store")
    parser.add_argument("--target", dest="target", action="store")
    parser.add_argument("--iface", dest="iface", action="store")
    parser.add_argument("--user", dest="user", action="store")
    parser.add_argument("--password", dest="password", action="store")
    parser.add_argument("--reverse-user", dest="reverse-user", action="store")
    parser.add_argument("--reverse-password", dest="reverse-password", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_iscsiname(rule):
    """
    Parse the iscsiname line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    # parser.add_argument('iqn')

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_keyboard(rule):
    """
    Parse the keyboard line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--vckeymap", dest="vckeymap", action="store")
    parser.add_argument("--xlayouts", dest="xlayouts", action="store")
    parser.add_argument("--switch", dest="switch", action="store")
    parser.add_argument("keyboard")

    args = clean_args(vars(parser.parse_args(rules)))

    if "keyboard" in args and "xlayouts" not in args:
        args["xlayouts"] = args["keyboard"]

    parser = None
    return args


def parse_lang(rule):
    """
    Parse the lang line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("lang")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_logvol(rule):
    """
    Parse the logvol line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("mntpoint")
    parser.add_argument("--noformat", dest="noformat", action="store_true")
    parser.add_argument("--useexisting", dest="useexisting", action="store_true")
    parser.add_argument("--fstype", dest="fstype", action="store")
    parser.add_argument("--fsoptions", dest="fsoptions", action="store")
    parser.add_argument("--grow", dest="grow", action="store_true")
    parser.add_argument("--maxsize", dest="maxsize", action="store")
    parser.add_argument("--recommended", dest="recommended", action="store_true")
    parser.add_argument("--percent", dest="percent", action="store_true")
    parser.add_argument("--encrypted", dest="encrypted", action="store_true")
    parser.add_argument("--passphrase", dest="passphrase", action="store")
    parser.add_argument("--escrowcert", dest="escrowcert", action="store")
    parser.add_argument(
        "--backuppassphrase", dest="backuppassphrase", action="store_true"
    )
    parser.add_argument("--name", dest="name", action="store")
    parser.add_argument("--vgname", dest="vgname", action="store")
    parser.add_argument("--size", dest="size", action="store")
    parser.add_argument("--label", dest="label", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_logging(rule):
    """
    Parse the logging line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--host", dest="host", action="store")
    parser.add_argument("--port", dest="port", action="store")
    parser.add_argument(
        "--level",
        dest="level",
        action="store",
        choices=["debug", "info", "warning", "error", "critical"],
    )

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_monitor(rule):
    """
    Parse the monitor line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--hsync", dest="hsync", action="store")
    parser.add_argument("--monitor", dest="monitor", action="store")
    parser.add_argument("--noprobe", dest="noprobe", action="store_true")
    parser.add_argument("--vsync", dest="vsync", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_multipath(rule):
    """
    Parse the multipath line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--name", dest="name", action="store")
    parser.add_argument("--device", dest="device", action="store")
    parser.add_argument("--rule", dest="rule", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_network(rule):
    """
    Parse the network line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument(
        "--bootproto",
        dest="bootproto",
        action="store",
        choices=["dhcp", "bootp", "static", "ibft"],
    )
    parser.add_argument("--device", dest="device", action="store")
    parser.add_argument("--ip", dest="ip", action="store")
    parser.add_argument("--ipv6", dest="ipv6", action="store")
    parser.add_argument("--gateway", dest="gateway", action="store")
    parser.add_argument("--nodefroute", dest="nodefroute", action="store_true")
    parser.add_argument("--nameserver", dest="nameserver", action="store")
    parser.add_argument("--nodns", dest="nodns", action="store_true")
    parser.add_argument("--netmask", dest="netmask", action="store")
    parser.add_argument("--hostname", dest="hostname", action="store")
    parser.add_argument("--ethtool", dest="ethtool", action="store")
    parser.add_argument("--essid", dest="essid", action="store")
    parser.add_argument("--wepkey", dest="wepkey", action="store")
    parser.add_argument("--wpakey", dest="wpakey", action="store")
    parser.add_argument("--onboot", dest="onboot", action="store")
    parser.add_argument("--dhcpclass", dest="dhcpclass", action="store")
    parser.add_argument("--mtu", dest="mtu", action="store")
    parser.add_argument("--noipv4", dest="noipv4", action="store_true")
    parser.add_argument("--noipv6", dest="noipv6", action="store_true")
    parser.add_argument("--activate", dest="activate", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_nfs(rule):
    """
    Parse the nfs line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--server", dest="server", action="store")
    parser.add_argument("--dir", dest="dir", action="store")
    parser.add_argument("--opts", dest="opts", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_partition(rule):
    """
    Parse the partition line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("mntpoint")
    parser.add_argument("--size", dest="size", action="store")
    parser.add_argument("--grow", dest="grow", action="store_true")
    parser.add_argument("--maxsize", dest="maxsize", action="store")
    parser.add_argument("--noformat", dest="noformat", action="store_true")
    parser.add_argument("--onpart", "--usepart", dest="onpart", action="store")
    parser.add_argument("--ondisk", "--ondrive", dest="ondisk", action="store")
    parser.add_argument("--asprimary", dest="asprimary", action="store_true")
    parser.add_argument("--fsprofile", dest="fsprofile", action="store")
    parser.add_argument("--fstype", dest="fstype", action="store")
    parser.add_argument("--fsoptions", dest="fsoptions", action="store")
    parser.add_argument("--label", dest="label", action="store")
    parser.add_argument("--recommended", dest="recommended", action="store_true")
    parser.add_argument("--onbiosdisk", dest="onbiosdisk", action="store")
    parser.add_argument("--encrypted", dest="encrypted", action="store_true")
    parser.add_argument("--passphrase", dest="passphrase", action="store")
    parser.add_argument("--escrowcert", dest="escrowcert", action="store")
    parser.add_argument("--backupphrase", dest="backupphrase", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_raid(rule):
    """
    Parse the raid line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)

    partitions = []
    newrules = []
    for count, rule in enumerate(rules):
        if count == 0:
            newrules.append(rule)
            continue
        elif rule.startswith("--"):
            newrules.append(rule)
            continue
        else:
            partitions.append(rule)
    rules = newrules

    parser.add_argument("mntpoint")
    parser.add_argument("--level", dest="level", action="store")
    parser.add_argument("--device", dest="device", action="store")
    parser.add_argument("--spares", dest="spares", action="store")
    parser.add_argument("--fstype", dest="fstype", action="store")
    parser.add_argument("--fsoptions", dest="fsoptions", action="store")
    parser.add_argument("--label", dest="label", action="store")
    parser.add_argument("--noformat", dest="noformat", action="store_true")
    parser.add_argument("--useexisting", dest="useexisting", action="store_true")
    parser.add_argument("--encrypted", dest="encrypted", action="store_true")
    parser.add_argument("--passphrase", dest="passphrase", action="store")
    parser.add_argument("--escrowcert", dest="escrowcert", action="store")
    parser.add_argument("--backuppassphrase", dest="backuppassphrase", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    if partitions:
        args["partitions"] = partitions
    parser = None
    return args


def parse_reboot(rule):
    """
    Parse the reboot line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--eject", dest="eject", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_repo(rule):
    """
    Parse the repo line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--name", dest="name", action="store")
    parser.add_argument("--baseurl", dest="baseurl", action="store")
    parser.add_argument("--mirrorlist", dest="mirrorlist", action="store")
    parser.add_argument("--cost", dest="cost", action="store")
    parser.add_argument("--excludepkgs", dest="excludepkgs", action="store")
    parser.add_argument("--includepkgs", dest="includepkgs", action="store")
    parser.add_argument("--proxy", dest="proxy", action="store")
    parser.add_argument("--ignoregroups", dest="ignoregroups", action="store")
    parser.add_argument("--noverifyssl", dest="noverifyssl", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_rescue(rule):
    """
    Parse the rescue line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--nomount", dest="nomount", action="store_true")
    parser.add_argument("--romount", dest="romount", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_rootpw(rule):
    """
    Parse the rootpw line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--iscrypted", dest="iscrypted", action="store_true")
    parser.add_argument("--plaintext", dest="plaintext", action="store_true")
    parser.add_argument("--lock", dest="lock", action="store_true")
    parser.add_argument("password")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_selinux(rule):
    """
    Parse the selinux line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--disabled", dest="disabled", action="store_true")
    parser.add_argument("--enforcing", dest="enforcing", action="store_true")
    parser.add_argument("--permissive", dest="permissive", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_services(rule):
    """
    Parse the services line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--disabled", dest="disabled", action="store")
    parser.add_argument("--enabled", dest="enabled", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_sshpw(rule):
    """
    Parse the sshpw line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--username", dest="username", action="store")
    parser.add_argument("--iscrypted", dest="iscrypted", action="store_true")
    parser.add_argument("--plaintext", dest="plaintext", action="store_true")
    parser.add_argument("--lock", dest="lock", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_timezone(rule):
    """
    Parse the timezone line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--utc", dest="utc", action="store_true")
    parser.add_argument("--nontp", dest="nontp", action="store_true")
    parser.add_argument("--ntpservers", dest="ntpservers", action="store")
    parser.add_argument("--isUtc", dest="isutc", action="store_true")
    parser.add_argument("timezone")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_updates(rule):
    """
    Parse the updates line
    """
    rules = shlex.split(rule)
    rules.pop(0)
    return {"url": rules[0]} if rules else True


def parse_upgrade(rule):
    """
    Parse the upgrade line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--root-device", dest="root-device", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    if args:
        return args
    return True


def parse_url(rule):
    """
    Parse the url line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--url", dest="url", action="store")
    parser.add_argument("--proxy", dest="proxy", action="store")
    parser.add_argument("--noverifyssl", dest="noverifyssl", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_user(rule):
    """
    Parse the user line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--name", dest="name", action="store")
    parser.add_argument("--gecos", dest="gecos", action="store")
    parser.add_argument("--groups", dest="groups", action="store")
    parser.add_argument("--homedir", dest="homedir", action="store")
    parser.add_argument("--lock", dest="lock", action="store_true")
    parser.add_argument("--password", dest="password", action="store")
    parser.add_argument("--iscrypted", dest="iscrypted", action="store_true")
    parser.add_argument("--plaintext", dest="plaintext", action="store_true")
    parser.add_argument("--shell", dest="shell", action="store")
    parser.add_argument("--uid", dest="uid", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_vnc(rule):
    """
    Parse the vnc line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--host", dest="host", action="store")
    parser.add_argument("--port", dest="port", action="store")
    parser.add_argument("--password", dest="password", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_volgroup(rule):
    """
    Parse the volgroup line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)

    partitions = []
    newrules = []
    for count, rule in enumerate(rules):
        if count == 0:
            newrules.append(rule)
            continue
        elif rule.startswith("--"):
            newrules.append(rule)
            continue
        else:
            partitions.append(rule)
    rules = newrules

    parser.add_argument("name")
    parser.add_argument("--noformat", dest="noformat", action="store_true")
    parser.add_argument("--useexisting", dest="useexisting", action="store_true")
    parser.add_argument("--pesize", dest="pesize", action="store")
    parser.add_argument("--reserved-space", dest="reserved-space", action="store")
    parser.add_argument("--reserved-percent", dest="reserved-percent", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    if partitions:
        args["partitions"] = partitions
    parser = None
    return args


def parse_xconfig(rule):
    """
    Parse the xconfig line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--defaultdesktop", dest="defaultdesktop", action="store")
    parser.add_argument("--startxonboot", dest="startxonboot", action="store_true")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def parse_zfcp(rule):
    """
    Parse the zfcp line
    """
    parser = argparse.ArgumentParser()
    rules = shlex.split(rule)
    rules.pop(0)
    parser.add_argument("--devnum", dest="devnum", action="store")
    parser.add_argument("--fcplun", dest="fcplun", action="store")
    parser.add_argument("--wwpn", dest="wwpn", action="store")

    args = clean_args(vars(parser.parse_args(rules)))
    parser = None
    return args


def mksls(src, dst=None):
    """
    Convert a kickstart file to an SLS file
    """
    mode = "command"
    sls = {}
    ks_opts = {}
    with salt.utils.files.fopen(src, "r") as fh_:
        for line in fh_:
            if line.startswith("#"):
                continue

            if mode == "command":
                if line.startswith("auth ") or line.startswith("authconfig "):
                    ks_opts["auth"] = parse_auth(line)
                elif line.startswith("autopart"):
                    ks_opts["autopath"] = parse_autopart(line)
                elif line.startswith("autostep"):
                    ks_opts["autostep"] = parse_autostep(line)
                elif line.startswith("bootloader"):
                    ks_opts["bootloader"] = parse_bootloader(line)
                elif line.startswith("btrfs"):
                    ks_opts["btrfs"] = parse_btrfs(line)
                elif line.startswith("cdrom"):
                    ks_opts["cdrom"] = True
                elif line.startswith("clearpart"):
                    ks_opts["clearpart"] = parse_clearpart(line)
                elif line.startswith("cmdline"):
                    ks_opts["cmdline"] = True
                elif line.startswith("device"):
                    ks_opts["device"] = parse_device(line)
                elif line.startswith("dmraid"):
                    ks_opts["dmraid"] = parse_dmraid(line)
                elif line.startswith("driverdisk"):
                    ks_opts["driverdisk"] = parse_driverdisk(line)
                elif line.startswith("firewall"):
                    ks_opts["firewall"] = parse_firewall(line)
                elif line.startswith("firstboot"):
                    ks_opts["firstboot"] = parse_firstboot(line)
                elif line.startswith("group"):
                    ks_opts["group"] = parse_group(line)
                elif line.startswith("graphical"):
                    ks_opts["graphical"] = True
                elif line.startswith("halt"):
                    ks_opts["halt"] = True
                elif line.startswith("harddrive"):
                    ks_opts["harddrive"] = True
                elif line.startswith("ignoredisk"):
                    ks_opts["ignoredisk"] = parse_ignoredisk(line)
                elif line.startswith("install"):
                    ks_opts["install"] = True
                elif line.startswith("iscsi"):
                    ks_opts["iscsi"] = parse_iscsi(line)
                elif line.startswith("iscsiname"):
                    ks_opts["iscsiname"] = parse_iscsiname(line)
                elif line.startswith("keyboard"):
                    ks_opts["keyboard"] = parse_keyboard(line)
                elif line.startswith("lang"):
                    ks_opts["lang"] = parse_lang(line)
                elif line.startswith("logvol"):
                    if "logvol" not in ks_opts:
                        ks_opts["logvol"] = []
                    ks_opts["logvol"].append(parse_logvol(line))
                elif line.startswith("logging"):
                    ks_opts["logging"] = parse_logging(line)
                elif line.startswith("mediacheck"):
                    ks_opts["mediacheck"] = True
                elif line.startswith("monitor"):
                    ks_opts["monitor"] = parse_monitor(line)
                elif line.startswith("multipath"):
                    ks_opts["multipath"] = parse_multipath(line)
                elif line.startswith("network"):
                    if "network" not in ks_opts:
                        ks_opts["network"] = []
                    ks_opts["network"].append(parse_network(line))
                elif line.startswith("nfs"):
                    ks_opts["nfs"] = True
                elif line.startswith("part ") or line.startswith("partition"):
                    if "part" not in ks_opts:
                        ks_opts["part"] = []
                    ks_opts["part"].append(parse_partition(line))
                elif line.startswith("poweroff"):
                    ks_opts["poweroff"] = True
                elif line.startswith("raid"):
                    if "raid" not in ks_opts:
                        ks_opts["raid"] = []
                    ks_opts["raid"].append(parse_raid(line))
                elif line.startswith("reboot"):
                    ks_opts["reboot"] = parse_reboot(line)
                elif line.startswith("repo"):
                    ks_opts["repo"] = parse_repo(line)
                elif line.startswith("rescue"):
                    ks_opts["rescue"] = parse_rescue(line)
                elif line.startswith("rootpw"):
                    ks_opts["rootpw"] = parse_rootpw(line)
                elif line.startswith("selinux"):
                    ks_opts["selinux"] = parse_selinux(line)
                elif line.startswith("services"):
                    ks_opts["services"] = parse_services(line)
                elif line.startswith("shutdown"):
                    ks_opts["shutdown"] = True
                elif line.startswith("sshpw"):
                    ks_opts["sshpw"] = parse_sshpw(line)
                elif line.startswith("skipx"):
                    ks_opts["skipx"] = True
                elif line.startswith("text"):
                    ks_opts["text"] = True
                elif line.startswith("timezone"):
                    ks_opts["timezone"] = parse_timezone(line)
                elif line.startswith("updates"):
                    ks_opts["updates"] = parse_updates(line)
                elif line.startswith("upgrade"):
                    ks_opts["upgrade"] = parse_upgrade(line)
                elif line.startswith("url"):
                    ks_opts["url"] = True
                elif line.startswith("user"):
                    ks_opts["user"] = parse_user(line)
                elif line.startswith("vnc"):
                    ks_opts["vnc"] = parse_vnc(line)
                elif line.startswith("volgroup"):
                    ks_opts["volgroup"] = parse_volgroup(line)
                elif line.startswith("xconfig"):
                    ks_opts["xconfig"] = parse_xconfig(line)
                elif line.startswith("zerombr"):
                    ks_opts["zerombr"] = True
                elif line.startswith("zfcp"):
                    ks_opts["zfcp"] = parse_zfcp(line)

            if line.startswith("%include"):
                rules = shlex.split(line)
                if not ks_opts["include"]:
                    ks_opts["include"] = []
                ks_opts["include"].append(rules[1])

            if line.startswith("%ksappend"):
                rules = shlex.split(line)
                if not ks_opts["ksappend"]:
                    ks_opts["ksappend"] = []
                ks_opts["ksappend"].append(rules[1])

            if line.startswith("%packages"):
                mode = "packages"
                if "packages" not in ks_opts:
                    ks_opts["packages"] = {"packages": {}}

                parser = argparse.ArgumentParser()
                opts = shlex.split(line)
                opts.pop(0)
                parser.add_argument("--default", dest="default", action="store_true")
                parser.add_argument(
                    "--excludedocs", dest="excludedocs", action="store_true"
                )
                parser.add_argument(
                    "--ignoremissing", dest="ignoremissing", action="store_true"
                )
                parser.add_argument("--instLangs", dest="instLangs", action="store")
                parser.add_argument("--multilib", dest="multilib", action="store_true")
                parser.add_argument(
                    "--nodefaults", dest="nodefaults", action="store_true"
                )
                parser.add_argument("--optional", dest="optional", action="store_true")
                parser.add_argument("--nobase", dest="nobase", action="store_true")
                args = clean_args(vars(parser.parse_args(opts)))
                ks_opts["packages"]["options"] = args

                continue

            if line.startswith("%pre"):
                mode = "pre"

                parser = argparse.ArgumentParser()
                opts = shlex.split(line)
                opts.pop(0)
                parser.add_argument("--interpreter", dest="interpreter", action="store")
                parser.add_argument(
                    "--erroronfail", dest="erroronfail", action="store_true"
                )
                parser.add_argument("--log", dest="log", action="store")
                args = clean_args(vars(parser.parse_args(opts)))
                ks_opts["pre"] = {"options": args, "script": ""}

                continue

            if line.startswith("%post"):
                mode = "post"

                parser = argparse.ArgumentParser()
                opts = shlex.split(line)
                opts.pop(0)
                parser.add_argument("--nochroot", dest="nochroot", action="store_true")
                parser.add_argument("--interpreter", dest="interpreter", action="store")
                parser.add_argument(
                    "--erroronfail", dest="erroronfail", action="store_true"
                )
                parser.add_argument("--log", dest="log", action="store")
                args = clean_args(vars(parser.parse_args(opts)))
                ks_opts["post"] = {"options": args, "script": ""}

                continue

            if line.startswith("%end"):
                mode = None

            if mode == "packages":
                if line.startswith("-"):
                    package = line.replace("-", "", 1).strip()
                    ks_opts["packages"]["packages"][package] = False
                else:
                    ks_opts["packages"]["packages"][line.strip()] = True

            if mode == "pre":
                ks_opts["pre"]["script"] += line

            if mode == "post":
                ks_opts["post"]["script"] += line

    # Set language
    sls[ks_opts["lang"]["lang"]] = {"locale": ["system"]}

    # Set keyboard
    sls[ks_opts["keyboard"]["xlayouts"]] = {"keyboard": ["system"]}

    # Set timezone
    sls[ks_opts["timezone"]["timezone"]] = {"timezone": ["system"]}
    if "utc" in ks_opts["timezone"]:
        sls[ks_opts["timezone"]["timezone"]]["timezone"].append("utc")

    # Set network
    if "network" in ks_opts:
        for interface in ks_opts["network"]:
            device = interface.get("device", None)
            if device is not None:
                del interface["device"]
                sls[device] = {"proto": interface["bootproto"]}
                del interface["bootproto"]

                if "onboot" in interface:
                    if "no" in interface["onboot"]:
                        sls[device]["enabled"] = False
                    else:
                        sls[device]["enabled"] = True
                    del interface["onboot"]

                if "noipv4" in interface:
                    sls[device]["ipv4"] = {"enabled": False}
                    del interface["noipv4"]
                if "noipv6" in interface:
                    sls[device]["ipv6"] = {"enabled": False}
                    del interface["noipv6"]

                for option in interface:
                    if type(interface[option]) is bool:
                        sls[device][option] = {"enabled": [interface[option]]}
                    else:
                        sls[device][option] = interface[option]
            if "hostname" in interface:
                sls["system"] = {
                    "network.system": {
                        "enabled": True,
                        "hostname": interface["hostname"],
                        "apply_hostname": True,
                    }
                }

    # Set selinux
    if "selinux" in ks_opts:
        for mode in ks_opts["selinux"]:
            sls[mode] = {"selinux": ["mode"]}

    # Get package data together
    if "nobase" not in ks_opts["packages"]["options"]:
        sls["base"] = {"pkg_group": ["installed"]}

    packages = ks_opts["packages"]["packages"]
    for package in packages:
        if not packages[package]:
            continue
        if package and packages[package] is True:
            if package.startswith("@"):
                pkg_group = package.replace("@", "", 1)
                sls[pkg_group] = {"pkg_group": ["installed"]}
            else:
                sls[package] = {"pkg": ["installed"]}
        elif packages[package] is False:
            sls[package] = {"pkg": ["absent"]}

    if dst:
        with salt.utils.files.fopen(dst, "w") as fp_:
            salt.utils.yaml.safe_dump(sls, fp_, default_flow_style=False)
    else:
        return salt.utils.yaml.safe_dump(sls, default_flow_style=False)



import datetime
import logging
import os
import signal
import subprocess
import textwrap

import pytest

import salt.utils.files

pytestmark = [
    pytest.mark.skip_unless_on_linux,
    pytest.mark.slow_test,
]

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def cmdmod(modules):
    return modules.cmd


@pytest.fixture(scope="module")
def grains(modules):
    return modules.grains


@pytest.fixture(scope="module")
def file(modules):
    return modules.file


@pytest.fixture(scope="module")
def service(modules):
    return modules.service


@pytest.fixture(scope="module")
def system(modules):
    return modules.system


@pytest.fixture(scope="function")
def fmt_str():
    return "%Y-%m-%d %H:%M:%S"


@pytest.fixture(scope="function")
def setup_teardown_vars(file, service, system):
    _orig_time = datetime.datetime.utcnow()

    if os.path.isfile("/etc/machine-info"):
        with salt.utils.files.fopen("/etc/machine-info", "r") as mach_info:
            _machine_info = mach_info.read()
    else:
        _machine_info = False

    try:
        _systemd_timesyncd_available_ = service.available("systemd-timesyncd")
        if _systemd_timesyncd_available_:
            res = service.stop("systemd-timesyncd")
            assert res

        yield _orig_time

    finally:
        _restore_time(system, _orig_time)

        if _machine_info is not False:
            with salt.utils.files.fopen("/etc/machine-info", "w") as mach_info:
                mach_info.write(_machine_info)
        else:
            file.remove("/etc/machine-info")

        if _systemd_timesyncd_available_:
            res = service.start("systemd-timesyncd")
            assert res


def _set_time(system, new_time, offset=None):
    t = new_time.timetuple()[:6]
    t += (offset,)
    return system.set_system_date_time(*t)


def _restore_time(system, _orig_time):
    result = _set_time(system, _orig_time, "+0000")
    assert result, "Unable to restore time properly"


def _same_times(t1, t2, seconds_diff=30):
    """
    Helper function to check if two datetime objects
    are close enough to the same time.
    """
    return abs(t1 - t2) < datetime.timedelta(seconds=seconds_diff)


@pytest.fixture(scope="module")
def hwclock_has_compare(cmdmod):
    """
    Some builds of hwclock don't include the `--compare` function
    needed to test hw/sw clock synchronization. Returns false on
    systems where it's not present so that we can skip the
    comparison portion of the test.
    """
    res = cmdmod.run_all(cmd="hwclock -h")
    _hwclock_has_compare_ = res["retcode"] == 0 and res["stdout"].find("--compare") > 0
    return _hwclock_has_compare_


def _test_hwclock_sync(system, hwclock_has_compare):
    """
    Check that hw and sw clocks are sync'd.
    """
    if not system.has_settable_hwclock():
        return None
    if not hwclock_has_compare:
        return None

    class CompareTimeout(BaseException):
        pass

    def _alrm_handler(sig, frame):
        log.warning("hwclock --compare failed to produce output after 3 seconds")
        raise CompareTimeout

    for _ in range(2):
        try:
            orig_handler = signal.signal(signal.SIGALRM, _alrm_handler)
            signal.alarm(3)
            rpipeFd, wpipeFd = os.pipe()
            log.debug("Comparing hwclock to sys clock")
            with os.fdopen(rpipeFd, "r") as rpipe:
                with os.fdopen(wpipeFd, "w"):
                    with salt.utils.files.fopen(os.devnull, "r") as nulFd:
                        p = subprocess.Popen(
                            args=["hwclock", "--compare"],
                            stdin=nulFd,
                            stdout=wpipeFd,
                            stderr=subprocess.PIPE,
                        )
                        p.communicate()

                        # read header
                        rpipe.readline()

                        # read first time comparison
                        timeCompStr = rpipe.readline()

                        # stop
                        p.terminate()

                        timeComp = timeCompStr.split()
                        hwTime = float(timeComp[0])
                        swTime = float(timeComp[1])
                        diff = abs(hwTime - swTime)

                        assert diff <= 2.0, "hwclock difference too big: " + str(
                            timeCompStr
                        )
                        break
        except CompareTimeout:
            p.terminate()
        finally:
            signal.alarm(0)
            signal.signal(signal.SIGALRM, orig_handler)
    else:
        log.error("Failed to check hwclock sync")


def test_get_system_date_time(setup_teardown_vars, system, fmt_str):
    """
    Test we are able to get the correct time
    """
    t1 = datetime.datetime.now()
    res = system.get_system_date_time()
    t2 = datetime.datetime.strptime(res, fmt_str)
    msg = "Difference in times is too large. Now: {} Fake: {}".format(t1, t2)
    assert _same_times(t1, t2, seconds_diff=3), msg


def test_get_system_date_time_utc(setup_teardown_vars, system, fmt_str):
    """
    Test we are able to get the correct time with utc
    """
    t1 = datetime.datetime.utcnow()
    res = system.get_system_date_time("+0000")
    t2 = datetime.datetime.strptime(res, fmt_str)
    msg = "Difference in times is too large. Now: {} Fake: {}".format(t1, t2)
    assert _same_times(t1, t2, seconds_diff=3), msg


@pytest.mark.skip_on_env("ON_DOCKER", eq="1")
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_set_system_date_time(setup_teardown_vars, system, hwclock_has_compare):
    """
    Test changing the system clock. We are only able to set it up to a
    resolution of a second so this test may appear to run in negative time.
    """
    cmp_time = datetime.datetime.now() - datetime.timedelta(days=7)
    result = _set_time(system, cmp_time)

    time_now = datetime.datetime.now()

    msg = "Difference in times is too large. Now: {} Fake: {}".format(
        time_now, cmp_time
    )
    assert result
    assert _same_times(time_now, cmp_time), msg
    _test_hwclock_sync(system, hwclock_has_compare)


@pytest.mark.skip_on_env("ON_DOCKER", eq="1")
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_set_system_date_time_utc(setup_teardown_vars, system, hwclock_has_compare):
    """
    Test changing the system clock. We are only able to set it up to a
    resolution of a second so this test may appear to run in negative time.
    """
    cmp_time = datetime.datetime.utcnow() - datetime.timedelta(days=7)
    result = _set_time(system, cmp_time, offset="+0000")

    time_now = datetime.datetime.utcnow()

    msg = "Difference in times is too large. Now: {} Fake: {}".format(
        time_now, cmp_time
    )
    assert result
    assert _same_times(time_now, cmp_time), msg
    _test_hwclock_sync(system, hwclock_has_compare)


@pytest.mark.skip_on_env("ON_DOCKER", eq="1")
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_set_system_date_time_utcoffset_east(
    setup_teardown_vars, system, hwclock_has_compare
):
    """
    Test changing the system clock. We are only able to set it up to a
    resolution of a second so this test may appear to run in negative time.
    """
    cmp_time = datetime.datetime.utcnow() - datetime.timedelta(days=7)
    # 25200 seconds = 7 hours
    time_to_set = cmp_time - datetime.timedelta(seconds=25200)
    result = _set_time(system, time_to_set, offset="-0700")
    time_now = datetime.datetime.utcnow()

    msg = "Difference in times is too large. Now: {} Fake: {}".format(
        time_now, cmp_time
    )
    assert result
    assert _same_times(time_now, cmp_time), msg
    _test_hwclock_sync(system, hwclock_has_compare)


@pytest.mark.skip_on_env("ON_DOCKER", eq="1")
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_set_system_date_time_utcoffset_west(
    setup_teardown_vars, system, hwclock_has_compare
):
    """
    Test changing the system clock. We are only able to set it up to a
    resolution of a second so this test may appear to run in negative time.
    """
    cmp_time = datetime.datetime.utcnow() - datetime.timedelta(days=7)
    # 7200 seconds = 2 hours
    time_to_set = cmp_time + datetime.timedelta(seconds=7200)
    result = _set_time(system, time_to_set, offset="+0200")
    time_now = datetime.datetime.utcnow()

    msg = "Difference in times is too large. Now: {} Fake: {}".format(
        time_now, cmp_time
    )
    assert result
    assert _same_times(time_now, cmp_time), msg
    _test_hwclock_sync(system, hwclock_has_compare)


@pytest.mark.skip_on_env("ON_DOCKER", eq="1")
@pytest.mark.flaky(max_runs=4)
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_set_system_time(setup_teardown_vars, system, hwclock_has_compare):
    """
    Test setting the system time without adjusting the date.
    """
    cmp_time = datetime.datetime.now().replace(hour=10, minute=5, second=0)

    result = system.set_system_time("10:05:00")

    time_now = datetime.datetime.now()
    msg = "Difference in times is too large. Now: {} Fake: {}".format(
        time_now, cmp_time
    )

    assert result
    assert _same_times(time_now, cmp_time), msg
    _test_hwclock_sync(system, hwclock_has_compare)


@pytest.mark.skip_on_env("ON_DOCKER", eq="1")
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_set_system_date(setup_teardown_vars, system, hwclock_has_compare):
    """
    Test setting the system date without adjusting the time.
    """
    cmp_time = datetime.datetime.now() - datetime.timedelta(days=7)

    result = system.set_system_date(cmp_time.strftime("%Y-%m-%d"))

    time_now = datetime.datetime.now()

    msg = "Difference in times is too large. Now: {} Fake: {}".format(
        time_now, cmp_time
    )

    assert result
    assert _same_times(time_now, cmp_time), msg
    _test_hwclock_sync(system, hwclock_has_compare)


@pytest.mark.skip_if_not_root
def test_get_computer_desc(setup_teardown_vars, system, cmdmod):
    """
    Test getting the system hostname
    """
    res = system.get_computer_desc()

    hostname_cmd = salt.utils.path.which("hostnamectl")
    if hostname_cmd:
        desc = cmdmod.run("hostnamectl status --pretty")
        assert res == desc
    else:
        if not os.path.isfile("/etc/machine-info"):
            assert not res
        else:
            with salt.utils.files.fopen("/etc/machine-info", "r") as mach_info:
                data = mach_info.read()
                assert res in data.decode("string_escape")


@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_set_computer_desc(setup_teardown_vars, system):
    """
    Test setting the computer description
    """
    desc = "test"
    ret = system.set_computer_desc(desc)
    computer_desc = system.get_computer_desc()

    assert ret
    assert desc in computer_desc


@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_set_computer_desc_multiline(setup_teardown_vars, system):
    """
    Test setting the computer description with a multiline string with tabs
    and double-quotes.
    """
    desc = textwrap.dedent(
        '''\
        'First Line
        \tSecond Line: 'single-quoted string'
        \t\tThird Line: "double-quoted string with unicode: питон"'''
    )
    ret = system.set_computer_desc(desc)
    computer_desc = salt.utils.stringutils.to_unicode(system.get_computer_desc())

    assert ret
    assert desc in computer_desc


@pytest.mark.skip_if_not_root
def test_has_hwclock(setup_teardown_vars, system, grains, hwclock_has_compare):
    """
    Verify platform has a settable hardware clock, if possible.
    """
    if grains.get("os_family") == "NILinuxRT":
        assert system._has_settable_hwclock()
        assert hwclock_has_compare(system)



import salt.ext.tornado
import salt.ext.tornado.testing
import salt.netapi.rest_tornado.saltnado as saltnado
from tests.support.mock import MagicMock, patch


class TestJobNotRunning(salt.ext.tornado.testing.AsyncTestCase):
    def setUp(self):
        super().setUp()
        self.mock = MagicMock()
        self.mock.opts = {
            "syndic_wait": 0.1,
            "cachedir": "/tmp/testing/cachedir",
            "sock_dir": "/tmp/testing/sock_drawer",
            "transport": "zeromq",
            "extension_modules": "/tmp/testing/moduuuuules",
            "order_masters": False,
            "gather_job_timeout": 10.001,
        }
        self.handler = saltnado.SaltAPIHandler(self.mock, self.mock)
        self.handler._write_buffer = []
        self.handler._transforms = []
        self.handler.lowstate = []
        self.handler.content_type = "text/plain"
        self.handler.dumper = lambda x: x
        f = salt.ext.tornado.gen.Future()
        f.set_result({"jid": f, "minions": []})
        self.handler.saltclients.update({"local": lambda *args, **kwargs: f})

    @salt.ext.tornado.testing.gen_test
    def test_when_disbatch_has_already_finished_then_writing_return_should_not_fail(
        self,
    ):
        self.handler.finish()
        result = yield self.handler.disbatch()
        # No assertion necessary, because we just want no failure here.
        # Asserting that it doesn't raise anything is... the default behavior
        # for a test.

    @salt.ext.tornado.testing.gen_test
    def test_when_disbatch_has_already_finished_then_finishing_should_not_fail(self):
        self.handler.finish()
        result = yield self.handler.disbatch()
        # No assertion necessary, because we just want no failure here.
        # Asserting that it doesn't raise anything is... the default behavior
        # for a test.

    @salt.ext.tornado.testing.gen_test
    def test_when_event_times_out_and_minion_is_not_running_result_should_be_True(self):
        fut = salt.ext.tornado.gen.Future()
        fut.set_exception(saltnado.TimeoutException())
        self.mock.event_listener.get_event.return_value = fut
        wrong_future = salt.ext.tornado.gen.Future()

        result = yield self.handler.job_not_running(
            jid=42, tgt="*", tgt_type="glob", minions=[], is_finished=wrong_future
        )

        self.assertTrue(result)

    @salt.ext.tornado.testing.gen_test
    def test_when_event_times_out_and_minion_is_not_running_minion_data_should_not_be_set(
        self,
    ):
        fut = salt.ext.tornado.gen.Future()
        fut.set_exception(saltnado.TimeoutException())
        self.mock.event_listener.get_event.return_value = fut
        wrong_future = salt.ext.tornado.gen.Future()
        minions = {}

        result = yield self.handler.job_not_running(
            jid=42, tgt="*", tgt_type="glob", minions=minions, is_finished=wrong_future
        )

        assert not minions

    @salt.ext.tornado.testing.gen_test
    def test_when_event_finally_finishes_and_returned_minion_not_in_minions_it_should_be_set_to_False(
        self,
    ):
        expected_id = 42
        no_data_event = salt.ext.tornado.gen.Future()
        no_data_event.set_result({"data": {}})
        empty_return_event = salt.ext.tornado.gen.Future()
        empty_return_event.set_result({"data": {"return": {}}})
        actual_return_event = salt.ext.tornado.gen.Future()
        actual_return_event.set_result(
            {"data": {"return": {"something happened here": "OK?"}, "id": expected_id}}
        )
        timed_out_event = salt.ext.tornado.gen.Future()
        timed_out_event.set_exception(saltnado.TimeoutException())
        self.mock.event_listener.get_event.side_effect = [
            no_data_event,
            empty_return_event,
            actual_return_event,
            timed_out_event,
            timed_out_event,
        ]
        minions = {}

        yield self.handler.job_not_running(
            jid=99,
            tgt="*",
            tgt_type="fnord",
            minions=minions,
            is_finished=salt.ext.tornado.gen.Future(),
        )

        self.assertFalse(minions[expected_id])

    @salt.ext.tornado.testing.gen_test
    def test_when_event_finally_finishes_and_returned_minion_already_in_minions_it_should_not_be_changed(
        self,
    ):
        expected_id = 42
        expected_value = object()
        minions = {expected_id: expected_value}
        no_data_event = salt.ext.tornado.gen.Future()
        no_data_event.set_result({"data": {}})
        empty_return_event = salt.ext.tornado.gen.Future()
        empty_return_event.set_result({"data": {"return": {}}})
        actual_return_event = salt.ext.tornado.gen.Future()
        actual_return_event.set_result(
            {"data": {"return": {"something happened here": "OK?"}, "id": expected_id}}
        )
        timed_out_event = salt.ext.tornado.gen.Future()
        timed_out_event.set_exception(saltnado.TimeoutException())
        self.mock.event_listener.get_event.side_effect = [
            no_data_event,
            empty_return_event,
            actual_return_event,
            timed_out_event,
            timed_out_event,
        ]

        yield self.handler.job_not_running(
            jid=99,
            tgt="*",
            tgt_type="fnord",
            minions=minions,
            is_finished=salt.ext.tornado.gen.Future(),
        )

        self.assertIs(minions[expected_id], expected_value)

    @salt.ext.tornado.testing.gen_test
    def test_when_event_returns_early_and_finally_times_out_result_should_be_True(self):
        no_data_event = salt.ext.tornado.gen.Future()
        no_data_event.set_result({"data": {}})
        empty_return_event = salt.ext.tornado.gen.Future()
        empty_return_event.set_result({"data": {"return": {}}})
        actual_return_event = salt.ext.tornado.gen.Future()
        actual_return_event.set_result(
            {"data": {"return": {"something happened here": "OK?"}, "id": "fnord"}}
        )
        timed_out_event = salt.ext.tornado.gen.Future()
        timed_out_event.set_exception(saltnado.TimeoutException())
        self.mock.event_listener.get_event.side_effect = [
            no_data_event,
            empty_return_event,
            actual_return_event,
            timed_out_event,
            timed_out_event,
        ]

        result = yield self.handler.job_not_running(
            jid=99,
            tgt="*",
            tgt_type="fnord",
            minions={},
            is_finished=salt.ext.tornado.gen.Future(),
        )
        self.assertTrue(result)

    @salt.ext.tornado.testing.gen_test
    def test_when_event_finishes_but_is_finished_is_done_then_result_should_be_True(
        self,
    ):
        expected_minion_id = "fnord"
        expected_minion_value = object()
        no_data_event = salt.ext.tornado.gen.Future()
        no_data_event.set_result({"data": {}})
        empty_return_event = salt.ext.tornado.gen.Future()
        empty_return_event.set_result({"data": {"return": {}}})
        actual_return_event = salt.ext.tornado.gen.Future()
        actual_return_event.set_result(
            {
                "data": {
                    "return": {"something happened here": "OK?"},
                    "id": expected_minion_id,
                }
            }
        )
        is_finished = salt.ext.tornado.gen.Future()

        def abort(*args, **kwargs):
            yield actual_return_event
            f = salt.ext.tornado.gen.Future()
            f.set_exception(saltnado.TimeoutException())
            is_finished.set_result("This is done")
            yield f
            assert False, "Never should make it here"

        minions = {expected_minion_id: expected_minion_value}

        self.mock.event_listener.get_event.side_effect = (x for x in abort())

        result = yield self.handler.job_not_running(
            jid=99,
            tgt="*",
            tgt_type="fnord",
            minions=minions,
            is_finished=is_finished,
        )
        self.assertTrue(result)

        # These are failsafes to ensure nothing super sideways happened
        self.assertTrue(len(minions) == 1, str(minions))
        self.assertIs(minions[expected_minion_id], expected_minion_value)

    @salt.ext.tornado.testing.gen_test
    def test_when_is_finished_times_out_before_event_finishes_result_should_be_True(
        self,
    ):
        # Other test times out with event - this one should time out for is_finished
        finished = salt.ext.tornado.gen.Future()
        finished.set_exception(saltnado.TimeoutException())
        wrong_future = salt.ext.tornado.gen.Future()
        self.mock.event_listener.get_event.return_value = wrong_future

        result = yield self.handler.job_not_running(
            jid=42, tgt="*", tgt_type="glob", minions=[], is_finished=finished
        )

        self.assertTrue(result)

    @salt.ext.tornado.testing.gen_test
    def test_when_is_finished_times_out_before_event_finishes_event_should_have_result_set_to_None(
        self,
    ):
        finished = salt.ext.tornado.gen.Future()
        finished.set_exception(saltnado.TimeoutException())
        wrong_future = salt.ext.tornado.gen.Future()
        self.mock.event_listener.get_event.return_value = wrong_future

        result = yield self.handler.job_not_running(
            jid=42, tgt="*", tgt_type="glob", minions=[], is_finished=finished
        )

        self.assertIsNone(wrong_future.result())


# TODO: I think we can extract seUp into a superclass -W. Werner, 2020-11-03
class TestGetMinionReturns(salt.ext.tornado.testing.AsyncTestCase):
    def setUp(self):
        super().setUp()
        self.mock = MagicMock()
        self.mock.opts = {
            "syndic_wait": 0.1,
            "cachedir": "/tmp/testing/cachedir",
            "sock_dir": "/tmp/testing/sock_drawer",
            "transport": "zeromq",
            "extension_modules": "/tmp/testing/moduuuuules",
            "order_masters": False,
            "gather_job_timeout": 10.001,
        }
        self.handler = saltnado.SaltAPIHandler(self.mock, self.mock)
        f = salt.ext.tornado.gen.Future()
        f.set_result({"jid": f, "minions": []})

    @salt.ext.tornado.testing.gen_test
    def test_if_finished_before_any_events_return_then_result_should_be_empty_dictionary(
        self,
    ):
        expected_result = {}
        xxx = salt.ext.tornado.gen.Future()
        xxx.set_result(None)
        is_finished = salt.ext.tornado.gen.Future()
        is_finished.set_result(None)
        actual_result = yield self.handler.get_minion_returns(
            events=[],
            is_finished=is_finished,
            is_timed_out=salt.ext.tornado.gen.Future(),
            min_wait_time=xxx,
            minions={},
        )
        self.assertDictEqual(actual_result, expected_result)

    # TODO: Copy above - test with timed out -W. Werner, 2020-11-05

    @salt.ext.tornado.testing.gen_test
    def test_if_is_finished_after_events_return_then_result_should_contain_event_result_data(
        self,
    ):
        expected_result = {
            "minion1": {"fnord": "this is some fnordish data"},
            "minion2": {"fnord": "this is some other fnordish data"},
        }
        xxx = salt.ext.tornado.gen.Future()
        xxx.set_result(None)
        is_finished = salt.ext.tornado.gen.Future()
        # XXX what do I do here?
        events = [
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
        ]
        events[0].set_result(
            {
                "tag": "fnord",
                "data": {"id": "minion1", "return": expected_result["minion1"]},
            }
        )
        events[1].set_result(
            {
                "tag": "fnord",
                "data": {"id": "minion2", "return": expected_result["minion2"]},
            }
        )
        self.io_loop.call_later(0.2, lambda: is_finished.set_result(None))

        actual_result = yield self.handler.get_minion_returns(
            events=events,
            is_finished=is_finished,
            is_timed_out=salt.ext.tornado.gen.Future(),
            min_wait_time=xxx,
            minions={
                "minion1": False,
                "minion2": False,
                "never returning minion": False,
            },
        )

        assert actual_result == expected_result

    @salt.ext.tornado.testing.gen_test
    def test_if_timed_out_after_events_return_then_result_should_contain_event_result_data(
        self,
    ):
        expected_result = {
            "minion1": {"fnord": "this is some fnordish data"},
            "minion2": {"fnord": "this is some other fnordish data"},
        }
        xxx = salt.ext.tornado.gen.Future()
        xxx.set_result(None)
        is_timed_out = salt.ext.tornado.gen.Future()
        # XXX what do I do here?
        events = [
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
        ]
        events[0].set_result(
            {
                "tag": "fnord",
                "data": {"id": "minion1", "return": expected_result["minion1"]},
            }
        )
        events[1].set_result(
            {
                "tag": "fnord",
                "data": {"id": "minion2", "return": expected_result["minion2"]},
            }
        )
        self.io_loop.call_later(0.2, lambda: is_timed_out.set_result(None))

        actual_result = yield self.handler.get_minion_returns(
            events=events,
            is_finished=salt.ext.tornado.gen.Future(),
            is_timed_out=is_timed_out,
            min_wait_time=xxx,
            minions={
                "minion1": False,
                "minion2": False,
                "never returning minion": False,
            },
        )

        assert actual_result == expected_result

    @salt.ext.tornado.testing.gen_test
    def test_if_wait_timer_is_not_done_even_though_results_are_then_data_should_not_yet_be_returned(
        self,
    ):
        expected_result = {
            "one": {"fnordy one": "one has some data"},
            "two": {"fnordy two": "two has some data"},
        }
        events = [salt.ext.tornado.gen.Future(), salt.ext.tornado.gen.Future()]
        events[0].set_result(
            {"tag": "fnord", "data": {"id": "one", "return": expected_result["one"]}}
        )
        events[1].set_result(
            {"tag": "fnord", "data": {"id": "two", "return": expected_result["two"]}}
        )
        wait_timer = salt.ext.tornado.gen.Future()
        fut = self.handler.get_minion_returns(
            events=events,
            is_finished=salt.ext.tornado.gen.Future(),
            is_timed_out=salt.ext.tornado.gen.Future(),
            min_wait_time=wait_timer,
            minions={"one": False, "two": False},
        )

        def boop():
            yield fut

        self.io_loop.spawn_callback(boop)
        yield salt.ext.tornado.gen.sleep(0.1)

        assert not fut.done()

        wait_timer.set_result(None)
        actual_result = yield fut

        assert actual_result == expected_result

    @salt.ext.tornado.testing.gen_test
    def test_when_is_finished_any_other_futures_should_be_canceled(self):
        events = [
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
        ]

        is_finished = salt.ext.tornado.gen.Future()
        is_finished.set_result(None)
        yield self.handler.get_minion_returns(
            events=events,
            is_finished=is_finished,
            is_timed_out=salt.ext.tornado.gen.Future(),
            min_wait_time=salt.ext.tornado.gen.Future(),
            minions={"one": False, "two": False},
        )

        are_done = [event.done() for event in events]
        assert all(are_done)

    @salt.ext.tornado.testing.gen_test
    def test_when_an_event_times_out_then_we_should_not_enter_an_infinite_loop(self):
        # NOTE: this test will enter an infinite loop if the code is broken. I
        # was not able to figure out a way to ensure that the test exits with
        # failure rather than stalling forever. That is because the
        # TimeoutException happens first and then tornado will never yield
        # control to another coroutine. Like a coroutine to remove the future
        # with the TimeoutException. It is also not possible to clear the
        # TimeoutException.

        events = [
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
        ]

        # Arguably any event would work, but 3 isn't the first, so it
        # gives us a little more confidence that this test is testing
        # correctly
        events[3].set_exception(saltnado.TimeoutException())
        times_out_later = salt.ext.tornado.gen.Future()
        # 0.5s should be long enough that the test gets through doing other
        # things before hitting this timeout, which will cancel all the
        # in-flight futures.
        self.io_loop.call_later(0.5, lambda: times_out_later.set_result(None))
        yield self.handler.get_minion_returns(
            events=events,
            is_finished=salt.ext.tornado.gen.Future(),
            is_timed_out=times_out_later,
            min_wait_time=salt.ext.tornado.gen.Future(),
            minions={"one": False, "two": False},
        )

        # Technically we don't /need/ to check that all events are done,
        # but it's incorrect to exit the function without ensuring all
        # futures are canceled.
        are_done = [event.done() for event in events]
        assert all(are_done)
        assert times_out_later.done()

    @salt.ext.tornado.testing.gen_test
    def test_when_is_timed_out_any_other_futures_should_be_canceled(self):
        # There is some question about whether this test is or should be
        # necessary. Or if it's meaningful. The code that this is testing
        # should never actually be able to make it to this point -- because
        # when all events have completed it should exit at a different branch.
        # That being said, the worst case is that this is just a duplicate
        # or irrelevant test, and can be removed.
        events = [
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
        ]

        is_timed_out = salt.ext.tornado.gen.Future()
        is_timed_out.set_result(None)
        yield self.handler.get_minion_returns(
            events=events,
            is_finished=salt.ext.tornado.gen.Future(),
            is_timed_out=is_timed_out,
            min_wait_time=salt.ext.tornado.gen.Future(),
            minions={"one": False, "two": False},
        )

        are_done = [event.done() for event in events]
        assert all(are_done)

    @salt.ext.tornado.testing.gen_test
    def test_when_min_wait_time_and_nothing_todo_any_other_futures_should_be_canceled(
        self,
    ):
        events = [
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
            salt.ext.tornado.gen.Future(),
        ]

        is_finished = salt.ext.tornado.gen.Future()
        min_wait_time = salt.ext.tornado.gen.Future()
        self.io_loop.call_later(0.2, lambda: min_wait_time.set_result(None))

        yield self.handler.get_minion_returns(
            events=events,
            is_finished=is_finished,
            is_timed_out=salt.ext.tornado.gen.Future(),
            min_wait_time=min_wait_time,
            minions={"one": True, "two": True},
        )

        are_done = [event.done() for event in events] + [is_finished.done()]
        assert all(are_done)

    @salt.ext.tornado.testing.gen_test
    def test_when_is_finished_but_not_is_timed_out_then_timed_out_should_not_be_set_to_done(
        self,
    ):
        events = [salt.ext.tornado.gen.Future()]
        is_timed_out = salt.ext.tornado.gen.Future()
        is_finished = salt.ext.tornado.gen.Future()
        is_finished.set_result(None)

        yield self.handler.get_minion_returns(
            events=events,
            is_finished=is_finished,
            is_timed_out=is_timed_out,
            min_wait_time=salt.ext.tornado.gen.Future(),
            minions={"one": False, "two": False},
        )

        assert not is_timed_out.done()

    @salt.ext.tornado.testing.gen_test
    def test_when_min_wait_time_and_all_completed_but_not_is_timed_out_then_timed_out_should_not_be_set_to_done(
        self,
    ):
        events = [salt.ext.tornado.gen.Future()]
        is_timed_out = salt.ext.tornado.gen.Future()
        min_wait_time = salt.ext.tornado.gen.Future()
        self.io_loop.call_later(0.2, lambda: min_wait_time.set_result(None))

        yield self.handler.get_minion_returns(
            events=events,
            is_finished=salt.ext.tornado.gen.Future(),
            is_timed_out=is_timed_out,
            min_wait_time=min_wait_time,
            minions={"one": True},
        )

        assert not is_timed_out.done()

    @salt.ext.tornado.testing.gen_test
    def test_when_things_are_completed_but_not_timed_out_then_timed_out_event_should_not_be_done(
        self,
    ):
        events = [
            salt.ext.tornado.gen.Future(),
        ]
        events[0].set_result({"tag": "fnord", "data": {"id": "one", "return": {}}})
        min_wait_time = salt.ext.tornado.gen.Future()
        min_wait_time.set_result(None)
        is_timed_out = salt.ext.tornado.gen.Future()

        yield self.handler.get_minion_returns(
            events=events,
            is_finished=salt.ext.tornado.gen.Future(),
            is_timed_out=is_timed_out,
            min_wait_time=min_wait_time,
            minions={"one": True},
        )

        assert not is_timed_out.done()


class TestDisbatchLocal(salt.ext.tornado.testing.AsyncTestCase):
    def setUp(self):
        super().setUp()
        self.mock = MagicMock()
        self.mock.opts = {
            "syndic_wait": 0.1,
            "cachedir": "/tmp/testing/cachedir",
            "sock_dir": "/tmp/testing/sock_drawer",
            "transport": "zeromq",
            "extension_modules": "/tmp/testing/moduuuuules",
            "order_masters": False,
            "gather_job_timeout": 10.001,
        }
        self.handler = saltnado.SaltAPIHandler(self.mock, self.mock)

    @salt.ext.tornado.testing.gen_test
    def test_when_is_timed_out_is_set_before_other_events_are_completed_then_result_should_be_empty_dictionary(
        self,
    ):
        completed_event = salt.ext.tornado.gen.Future()
        never_completed = salt.ext.tornado.gen.Future()
        # TODO: We may need to tweak these values to get them close enough but not so far away -W. Werner, 2020-11-17
        gather_timeout = 0.1
        event_timeout = gather_timeout + 0.05

        def fancy_get_event(*args, **kwargs):
            if kwargs.get("tag").endswith("/ret"):
                return never_completed
            return completed_event

        def completer():
            completed_event.set_result(
                {
                    "tag": "fnord",
                    "data": {
                        "return": "This should never be in chunk_ret",
                        "id": "fnord",
                    },
                }
            )

        self.io_loop.call_later(event_timeout, completer)

        f = salt.ext.tornado.gen.Future()
        f.set_result({"jid": "42", "minions": []})
        with patch.object(
            self.handler.application.event_listener,
            "get_event",
            side_effect=fancy_get_event,
        ), patch.dict(
            self.handler.application.opts,
            {"gather_job_timeout": gather_timeout, "timeout": 42},
        ), patch.dict(
            self.handler.saltclients, {"local": lambda *args, **kwargs: f}
        ):
            result = yield self.handler._disbatch_local(
                chunk={"tgt": "*", "tgt_type": "glob", "fun": "test.ping"}
            )

        assert result == {}

    @salt.ext.tornado.testing.gen_test
    def test_when_is_finished_is_set_before_events_return_then_no_data_should_be_returned(
        self,
    ):
        completed_event = salt.ext.tornado.gen.Future()
        never_completed = salt.ext.tornado.gen.Future()
        gather_timeout = 2
        event_timeout = gather_timeout - 1

        def fancy_get_event(*args, **kwargs):
            if kwargs.get("tag").endswith("/ret"):
                return never_completed
            return completed_event

        def completer():
            completed_event.set_result(
                {
                    "tag": "fnord",
                    "data": {
                        "return": "This should never be in chunk_ret",
                        "id": "fnord",
                    },
                }
            )

        self.io_loop.call_later(event_timeout, completer)

        def toggle_is_finished(*args, **kwargs):
            finished = kwargs.get("is_finished", args[4] if len(args) > 4 else None)
            assert finished is not None
            finished.set_result(42)

        f = salt.ext.tornado.gen.Future()
        f.set_result({"jid": "42", "minions": []})
        with patch.object(
            self.handler.application.event_listener,
            "get_event",
            side_effect=fancy_get_event,
        ), patch.object(
            self.handler,
            "job_not_running",
            autospec=True,
            side_effect=toggle_is_finished,
        ), patch.dict(
            self.handler.application.opts,
            {"gather_job_timeout": gather_timeout, "timeout": 42},
        ), patch.dict(
            self.handler.saltclients, {"local": lambda *args, **kwargs: f}
        ):
            result = yield self.handler._disbatch_local(
                chunk={"tgt": "*", "tgt_type": "glob", "fun": "test.ping"}
            )

        assert result == {}

    @salt.ext.tornado.testing.gen_test
    def test_when_is_finished_then_all_collected_data_should_be_returned(self):
        completed_event = salt.ext.tornado.gen.Future()
        never_completed = salt.ext.tornado.gen.Future()
        # This timeout should never be reached
        gather_timeout = 42
        completed_events = [salt.ext.tornado.gen.Future() for _ in range(5)]
        for i, event in enumerate(completed_events):
            event.set_result(
                {
                    "tag": "fnord",
                    "data": {
                        "return": f"return from fnord {i}",
                        "id": f"fnord {i}",
                    },
                }
            )
        uncompleted_events = [salt.ext.tornado.gen.Future() for _ in range(5)]
        events = iter(completed_events + uncompleted_events)
        expected_result = {
            "fnord 0": "return from fnord 0",
            "fnord 1": "return from fnord 1",
            "fnord 2": "return from fnord 2",
            "fnord 3": "return from fnord 3",
            "fnord 4": "return from fnord 4",
        }

        def fancy_get_event(*args, **kwargs):
            if kwargs.get("tag").endswith("/ret"):
                return never_completed
            else:
                return next(events)

        def toggle_is_finished(*args, **kwargs):
            finished = kwargs.get("is_finished", args[4] if len(args) > 4 else None)
            assert finished is not None
            finished.set_result(42)

        f = salt.ext.tornado.gen.Future()
        f.set_result({"jid": "42", "minions": ["non-existent minion"]})
        with patch.object(
            self.handler.application.event_listener,
            "get_event",
            side_effect=fancy_get_event,
        ), patch.object(
            self.handler,
            "job_not_running",
            autospec=True,
            side_effect=toggle_is_finished,
        ), patch.dict(
            self.handler.application.opts,
            {"gather_job_timeout": gather_timeout, "timeout": 42},
        ), patch.dict(
            self.handler.saltclients, {"local": lambda *args, **kwargs: f}
        ):
            result = yield self.handler._disbatch_local(
                chunk={"tgt": "*", "tgt_type": "glob", "fun": "test.ping"}
            )

        assert result == expected_result

    @salt.ext.tornado.testing.gen_test
    def test_when_is_timed_out_then_all_collected_data_should_be_returned(self):
        completed_event = salt.ext.tornado.gen.Future()
        never_completed = salt.ext.tornado.gen.Future()
        # 2s is probably enough for any kind of computer to manage to
        # do all the other processing. We could maybe reduce this - just
        # depends on how slow of a system we're running on.
        # TODO: Maybe we should have a test helper/fixture that benchmarks the system and gets a reasonable timeout? -W. Werner, 2020-11-19
        gather_timeout = 2
        completed_events = [salt.ext.tornado.gen.Future() for _ in range(5)]
        for i, event in enumerate(completed_events):
            event.set_result(
                {
                    "tag": "fnord",
                    "data": {
                        "return": f"return from fnord {i}",
                        "id": f"fnord {i}",
                    },
                }
            )
        uncompleted_events = [salt.ext.tornado.gen.Future() for _ in range(5)]
        events = iter(completed_events + uncompleted_events)
        expected_result = {
            "fnord 0": "return from fnord 0",
            "fnord 1": "return from fnord 1",
            "fnord 2": "return from fnord 2",
            "fnord 3": "return from fnord 3",
            "fnord 4": "return from fnord 4",
        }

        def fancy_get_event(*args, **kwargs):
            if kwargs.get("tag").endswith("/ret"):
                return never_completed
            else:
                return next(events)

        f = salt.ext.tornado.gen.Future()
        f.set_result({"jid": "42", "minions": ["non-existent minion"]})
        with patch.object(
            self.handler.application.event_listener,
            "get_event",
            side_effect=fancy_get_event,
        ), patch.dict(
            self.handler.application.opts,
            {"gather_job_timeout": gather_timeout, "timeout": 42},
        ), patch.dict(
            self.handler.saltclients, {"local": lambda *args, **kwargs: f}
        ):
            result = yield self.handler._disbatch_local(
                chunk={"tgt": "*", "tgt_type": "glob", "fun": "test.ping"}
            )

        assert result == expected_result

    @salt.ext.tornado.testing.gen_test
    def test_when_minions_all_return_then_all_collected_data_should_be_returned(self):
        completed_event = salt.ext.tornado.gen.Future()
        never_completed = salt.ext.tornado.gen.Future()
        # Timeout is something ridiculously high - it should never be reached
        gather_timeout = 20
        completed_events = [salt.ext.tornado.gen.Future() for _ in range(10)]
        events_by_id = {}
        for i, event in enumerate(completed_events):
            id_ = f"fnord {i}"
            events_by_id[id_] = event
            event.set_result(
                {
                    "tag": "fnord",
                    "data": {"return": f"return from {id_}", "id": id_},
                }
            )
        expected_result = {
            "fnord 0": "return from fnord 0",
            "fnord 1": "return from fnord 1",
            "fnord 2": "return from fnord 2",
            "fnord 3": "return from fnord 3",
            "fnord 4": "return from fnord 4",
            "fnord 5": "return from fnord 5",
            "fnord 6": "return from fnord 6",
            "fnord 7": "return from fnord 7",
            "fnord 8": "return from fnord 8",
            "fnord 9": "return from fnord 9",
        }

        def fancy_get_event(*args, **kwargs):
            tag = kwargs.get("tag", "").rpartition("/")[-1]
            return events_by_id.get(tag, never_completed)

        f = salt.ext.tornado.gen.Future()
        f.set_result(
            {
                "jid": "42",
                "minions": [e.result()["data"]["id"] for e in completed_events],
            }
        )
        with patch.object(
            self.handler.application.event_listener,
            "get_event",
            side_effect=fancy_get_event,
        ), patch.dict(
            self.handler.application.opts,
            {"gather_job_timeout": gather_timeout, "timeout": 42},
        ), patch.dict(
            self.handler.saltclients, {"local": lambda *args, **kwargs: f}
        ):
            result = yield self.handler._disbatch_local(
                chunk={"tgt": "*", "tgt_type": "glob", "fun": "test.ping"}
            )

        assert result == expected_result

    @salt.ext.tornado.testing.gen_test
    def test_when_min_wait_time_has_not_passed_then_disbatch_should_not_return_expected_data_until_time_has_passed(
        self,
    ):
        completed_event = salt.ext.tornado.gen.Future()
        never_completed = salt.ext.tornado.gen.Future()
        wait_timer = salt.ext.tornado.gen.Future()
        gather_timeout = 20
        completed_events = [salt.ext.tornado.gen.Future() for _ in range(10)]
        events_by_id = {}
        # Setup some real-enough looking return data
        for i, event in enumerate(completed_events):
            id_ = f"fnord {i}"
            events_by_id[id_] = event
            event.set_result(
                {
                    "tag": "fnord",
                    "data": {"return": f"return from {id_}", "id": id_},
                }
            )
        # Hard coded instead of dynamic to avoid potentially writing a test
        # that does nothing
        expected_result = {
            "fnord 0": "return from fnord 0",
            "fnord 1": "return from fnord 1",
            "fnord 2": "return from fnord 2",
            "fnord 3": "return from fnord 3",
            "fnord 4": "return from fnord 4",
            "fnord 5": "return from fnord 5",
            "fnord 6": "return from fnord 6",
            "fnord 7": "return from fnord 7",
            "fnord 8": "return from fnord 8",
            "fnord 9": "return from fnord 9",
        }

        # If this is one of our fnord events, return that future, otherwise
        # they're bogus events that are irrelevant to our current testing.
        # They get to wait for-ev-errrrr
        def fancy_get_event(*args, **kwargs):
            tag = kwargs.get("tag", "").rpartition("/")[-1]
            return events_by_id.get(tag, never_completed)

        minions = {}

        def capture_minions(*args, **kwargs):
            """
            Take minions that would be passed to a function, and
            store them for later checking.
            """
            nonlocal minions
            minions = args[3]

        # Needed to have both a fake sleep, as well as a *real* sleep.
        # The fake sleep is necessary so that we can return our own
        # min_wait_time future. The fakeo_timer object is how we signal
        # which one we need to be returning.
        orig_sleep = salt.ext.tornado.gen.sleep

        fakeo_timer = object()

        @salt.ext.tornado.gen.coroutine
        def fake_sleep(timer):
            # only return our fake min_wait_time future when the sentinel
            # value is provided. Otherwise it's just a number.
            if timer is fakeo_timer:
                yield wait_timer
            else:
                yield orig_sleep(timer)

        f = salt.ext.tornado.gen.Future()
        f.set_result(
            {
                "jid": "42",
                "minions": [e.result()["data"]["id"] for e in completed_events],
            }
        )
        with patch.object(
            self.handler.application.event_listener,
            "get_event",
            side_effect=fancy_get_event,
        ), patch.object(
            self.handler,
            "job_not_running",
            autospec=True,
            side_effect=capture_minions,
        ), patch.dict(
            self.handler.application.opts,
            {
                "gather_job_timeout": gather_timeout,
                "timeout": 42,
                "syndic_wait": fakeo_timer,
                "order_masters": True,
            },
        ), patch(
            "salt.ext.tornado.gen.sleep",
            autospec=True,
            side_effect=fake_sleep,
        ), patch.dict(
            self.handler.saltclients, {"local": lambda *args, **kwargs: f}
        ):

            # Example timeline that we're testing:
            #
            # If there's a min wait time of 10s, and all the results come
            # back in 5s, we still need to wait the full 10s.
            #
            # Here:
            # t=0, all events are completed
            # t=0.1, we check that all minions have been set to True, i.e. all
            #        events are completed. We also ensure that the future has
            #        not completed.
            # t=0.1+, we complete our injected timer, and then ensure that all
            #         the correct data has been returned.

            fut = self.handler._disbatch_local(
                chunk={"tgt": "*", "tgt_type": "glob", "fun": "test.ping"}
            )

            def boop():
                yield fut

            self.io_loop.spawn_callback(boop)
            yield salt.ext.tornado.gen.sleep(0.1)
            # here, all the minions should be complete (i.e. "True")
            assert all(minions[m_id] for m_id in minions)
            # But _disbatch_local is not returned yet because min_wait_time has not passed
            assert not fut.done()
            wait_timer.set_result(None)
            result = yield fut

        assert result == expected_result

    # Question: Currently, job_not_running can add to the minions dict, which
    # affects the more_todo result. However, the events are never added to
    # once we have entered the loop. I'm not sure if this is an oversight, or
    # simply an implicit expectation. I am making the assumption that this
    # behavior is correct and does not need extra testing. Otherwise, we should
    # be testing that when minions are added within job_not_running, that it
    # should affect the regular loop
    # -W. Werner, 2020-11-19



import pytest
from saltfactories.utils import random_string


@pytest.fixture
def salt_master(salt_factories):
    config_defaults = {
        "transport": "zeromq",
        "auto_accept": True,
        "sign_pub_messages": False,
    }
    factory = salt_factories.salt_master_daemon(
        random_string("zeromq-master-"), defaults=config_defaults
    )
    return factory


@pytest.fixture
def salt_minion(salt_master):
    config_defaults = {
        "transport": "zeromq",
        "master_ip": "127.0.0.1",
        "master_port": salt_master.config["ret_port"],
        "auth_timeout": 5,
        "auth_tries": 1,
        "master_uri": "tcp://127.0.0.1:{}".format(salt_master.config["ret_port"]),
    }
    factory = salt_master.salt_minion_daemon(
        random_string("zeromq-minion-"), defaults=config_defaults
    )
    return factory



import logging
import os

import pytest

import salt.states.file as filestate
import salt.utils.platform
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {filestate: {"__salt__": {}, "__opts__": {}}}


@pytest.fixture
def directory_name():
    name = os.sep + "test"
    if salt.utils.platform.is_windows():
        name = "c:" + name
    return name


def test_pruned_clean(directory_name):
    with patch("os.path.isdir", return_value=False):
        ret = filestate.pruned(name=directory_name)
    assert ret == {
        "changes": {},
        "comment": "Directory {} is not present".format(directory_name),
        "name": directory_name,
        "result": True,
    }


def test_pruned_test(directory_name):
    with patch("os.path.isdir", return_value=True), patch.dict(
        filestate.__opts__, {"test": True}
    ):
        ret = filestate.pruned(name=directory_name)
    assert ret == {
        "changes": {"deleted": directory_name},
        "comment": "Directory {} is set for removal".format(directory_name),
        "name": directory_name,
        "result": None,
    }


def test_pruned_success(directory_name):
    rmdir = MagicMock(return_value={"result": True})
    with patch("os.path.isdir", return_value=True), patch.dict(
        filestate.__opts__, {"test": False}
    ), patch.dict(filestate.__salt__, {"file.rmdir": rmdir}):
        ret = filestate.pruned(name=directory_name)
    assert ret == {
        "changes": {"deleted": directory_name},
        "comment": "Removed directory {}".format(directory_name),
        "name": directory_name,
        "result": True,
    }



"""
    :codeauthor: :email:`Shane Lee <slee@saltstack.com>`
"""

import errno
import textwrap

import salt.grains.iscsi as iscsi
from tests.support.mock import MagicMock, mock_open, patch


def test_windows_iscsi_iqn_grains():
    cmd_run_mock = MagicMock(
        return_value={"stdout": "iSCSINodeName\niqn.1991-05.com.microsoft:simon-x1\n"}
    )
    _grains = {}
    with patch("salt.utils.path.which", MagicMock(return_value=True)):
        with patch("salt.modules.cmdmod.run_all", cmd_run_mock):
            _grains["iscsi_iqn"] = iscsi._windows_iqn()

    assert _grains.get("iscsi_iqn") == ["iqn.1991-05.com.microsoft:simon-x1"]


def test_aix_iscsi_iqn_grains():
    cmd_run_mock = MagicMock(
        return_value="initiator_name iqn.localhost.hostid.7f000001"
    )

    _grains = {}
    with patch("salt.modules.cmdmod.run", cmd_run_mock):
        _grains["iscsi_iqn"] = iscsi._aix_iqn()

    assert _grains.get("iscsi_iqn") == ["iqn.localhost.hostid.7f000001"]


def test_linux_iscsi_iqn_grains():
    _iscsi_file = textwrap.dedent(
        """\
        ## DO NOT EDIT OR REMOVE THIS FILE!
        ## If you remove this file, the iSCSI daemon will not start.
        ## If you change the InitiatorName, existing access control lists
        ## may reject this initiator.  The InitiatorName must be unique
        ## for each iSCSI initiator.  Do NOT duplicate iSCSI InitiatorNames.
        InitiatorName=iqn.1993-08.org.debian:01:d12f7aba36
        """
    )

    with patch("salt.utils.files.fopen", mock_open(read_data=_iscsi_file)):
        iqn = iscsi._linux_iqn()

    assert isinstance(iqn, list)
    assert len(iqn) == 1
    assert iqn == ["iqn.1993-08.org.debian:01:d12f7aba36"]


def test_linux_iqn_non_root():
    """
    Test if linux_iqn is running on salt-master as non-root
    and handling access denial properly.
    :return:
    """
    with patch(
        "salt.utils.files.fopen",
        side_effect=IOError(errno.EPERM, "The cables are not the same length."),
    ):
        with patch("salt.grains.iscsi.log"):
            assert iscsi._linux_iqn() == []
            iscsi.log.debug.assert_called()
            assert "Error while accessing" in iscsi.log.debug.call_args[0][0]
            assert "cables are not the same" in iscsi.log.debug.call_args[0][2].strerror
            assert iscsi.log.debug.call_args[0][2].errno == errno.EPERM
            assert iscsi.log.debug.call_args[0][1] == "/etc/iscsi/initiatorname.iscsi"


def test_linux_iqn_no_iscsii_initiator():
    """
    Test if linux_iqn is running on salt-master as root.
    iscsii initiator is not there accessible or is not supported.
    :return:
    """
    with patch("salt.utils.files.fopen", side_effect=IOError(errno.ENOENT, "")):
        with patch("salt.grains.iscsi.log"):
            assert iscsi._linux_iqn() == []
            iscsi.log.debug.assert_not_called()



import pytest
import zmq

import salt.config
import salt.ext.tornado.ioloop
import salt.utils.event
import salt.utils.stringutils

NO_LONG_IPC = False
if getattr(zmq, "IPC_PATH_MAX_LEN", 103) <= 103:
    NO_LONG_IPC = True


pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skipif(
        NO_LONG_IPC,
        reason="This system does not support long IPC paths. Skipping event tests!",
    ),
]


class IOLoopContainer:
    def stop(self):
        self.io_loop.stop()
        self.publisher.close()
        self.event.destroy()

    def _handle_publish(self, raw):
        self.tag, self.data = salt.utils.event.SaltEvent.unpack(raw)
        self.stop()


@pytest.fixture()
def io_cont(io_loop, tmp_path):
    io_cont = IOLoopContainer()
    io_cont.io_loop = io_loop
    socks_dir_path = tmp_path / "test-socks"
    io_cont.sock_dir = str(socks_dir_path)
    socks_dir_path.mkdir(exist_ok=True)
    io_cont.opts = {"sock_dir": io_cont.sock_dir}
    io_cont.publisher = salt.utils.event.AsyncEventPublisher(
        io_cont.opts,
        io_cont.io_loop,
    )
    io_cont.event = salt.utils.event.get_event(
        "minion", opts=io_cont.opts, io_loop=io_cont.io_loop
    )
    io_cont.event.subscribe("")
    io_cont.event.set_event_handler(io_cont._handle_publish)
    yield io_cont


def test_event_subscription(io_cont):
    """Test a single event is received"""
    with salt.utils.event.MinionEvent(io_cont.opts, listen=True) as me:
        me.fire_event({"data": "foo1"}, "evt1")
        # Make sure the io_loop stops, always
        io_cont.io_loop.add_timeout(io_cont.io_loop.time() + 5, io_cont.io_loop.stop)
        # _handle_publish should stop the io_loop
        io_cont.io_loop.start()
        evt1 = me.get_event(tag="evt1")
        assert io_cont.tag == "evt1"
        io_cont.data.pop("_stamp")  # drop the stamp
        assert io_cont.data == {"data": "foo1"}


def test_event_unsubscribe_remove_error(io_cont):
    with salt.utils.event.MinionEvent(io_cont.opts, listen=True) as me:
        tag = "evt1"
        me.fire_event({"data": "foo1"}, tag)

        # Make sure no remove error is raised when tag is not found
        for _ in range(2):
            me.unsubscribe(tag)

        me.unsubscribe("tag_does_not_exist")



"""
Detect LVM Volumes
"""

import logging

import salt.modules.cmdmod
import salt.utils.files
import salt.utils.path
import salt.utils.platform

__salt__ = {
    "cmd.run": salt.modules.cmdmod._run_quiet,
    "cmd.run_all": salt.modules.cmdmod._run_all_quiet,
}

log = logging.getLogger(__name__)


def lvm():
    """
    Return list of LVM devices
    """
    if salt.utils.platform.is_linux():
        return _linux_lvm()
    elif salt.utils.platform.is_aix():
        return _aix_lvm()
    else:
        log.trace("LVM grain does not support this OS")


def _linux_lvm():
    ret = {}
    cmd = salt.utils.path.which("lvm")
    if cmd:
        vgs = __salt__["cmd.run_all"]("{} vgs -o vg_name --noheadings".format(cmd))

        for vg in vgs["stdout"].splitlines():
            vg = vg.strip()
            ret[vg] = []
            lvs = __salt__["cmd.run_all"](
                "{} lvs -o lv_name --noheadings {}".format(cmd, vg)
            )
            for lv in lvs["stdout"].splitlines():
                ret[vg].append(lv.strip())

        return {"lvm": ret}
    else:
        log.trace("No LVM installed")


def _aix_lvm():
    ret = {}
    cmd = salt.utils.path.which("lsvg")
    vgs = __salt__["cmd.run"]("{}".format(cmd))

    for vg in vgs.splitlines():
        ret[vg] = []
        lvs = __salt__["cmd.run"]("{} -l {}".format(cmd, vg))
        for lvline in lvs.splitlines()[2:]:
            lv = lvline.split(" ", 1)[0]
            ret[vg].append(lv)

    return {"lvm": ret}



"""
Tests for salt.utils.jid
"""


import datetime
import os

import salt.utils.jid
from tests.support.mock import patch
from tests.support.unit import TestCase


class JidTestCase(TestCase):
    def test_jid_to_time(self):
        test_jid = 20131219110700123489
        expected_jid = "2013, Dec 19 11:07:00.123489"
        self.assertEqual(salt.utils.jid.jid_to_time(test_jid), expected_jid)

        # Test incorrect lengths
        incorrect_jid_length = 2012
        self.assertEqual(salt.utils.jid.jid_to_time(incorrect_jid_length), "")

    def test_is_jid(self):
        self.assertTrue(salt.utils.jid.is_jid("20131219110700123489"))  # Valid JID
        self.assertFalse(salt.utils.jid.is_jid(20131219110700123489))  # int
        self.assertFalse(
            salt.utils.jid.is_jid("2013121911070012348911111")
        )  # Wrong length

    def test_gen_jid(self):
        now = datetime.datetime(2002, 12, 25, 12, 0, 0, 0)
        with patch("salt.utils.jid._utc_now", return_value=now):
            ret = salt.utils.jid.gen_jid({})
            self.assertEqual(ret, "20021225120000000000")
            with patch("salt.utils.jid.LAST_JID_DATETIME", None):
                ret = salt.utils.jid.gen_jid({"unique_jid": True})
                self.assertEqual(ret, "20021225120000000000_{}".format(os.getpid()))
                ret = salt.utils.jid.gen_jid({"unique_jid": True})
                self.assertEqual(ret, "20021225120000000001_{}".format(os.getpid()))

    def test_deprecation_58225(self):
        # check that type error will be raised
        self.assertRaises(TypeError, salt.utils.jid.gen_jid)

        # check that opts is missing and not another arg
        try:
            salt.utils.jid.gen_jid()  # pylint: disable=no-value-for-parameter
        except TypeError as no_opts:
            self.assertEqual(
                str(no_opts), "gen_jid() missing 1 required positional argument: 'opts'"
            )



"""
    Test cases for salt.modules.win_licence
"""


import pytest

import salt.modules.win_license as win_license
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_license: {}}


def test_installed():
    """
    Test to see if the given license key is installed
    """
    mock = MagicMock(return_value="Partial Product Key: ABCDE")
    with patch.dict(win_license.__salt__, {"cmd.run": mock}):
        out = win_license.installed("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
        mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /dli")
        assert out


def test_installed_diff():
    """
    Test to see if the given license key is installed when the key is different
    """
    mock = MagicMock(return_value="Partial Product Key: 12345")
    with patch.dict(win_license.__salt__, {"cmd.run": mock}):
        out = win_license.installed("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
        mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /dli")
        assert not out


def test_install():
    """
    Test installing the given product key
    """
    mock = MagicMock()
    with patch.dict(win_license.__salt__, {"cmd.run": mock}):
        win_license.install("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
        mock.assert_called_once_with(
            r"cscript C:\Windows\System32\slmgr.vbs /ipk "
            "AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE"
        )


def test_uninstall():
    """
    Test uninstalling the given product key
    """
    mock = MagicMock()
    with patch.dict(win_license.__salt__, {"cmd.run": mock}):
        win_license.uninstall()
        mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /upk")


def test_activate():
    """
    Test activating the current product key
    """
    mock = MagicMock()
    with patch.dict(win_license.__salt__, {"cmd.run": mock}):
        win_license.activate()
        mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /ato")


def test_licensed():
    """
    Test checking if the minion is licensed
    """
    mock = MagicMock(return_value="License Status: Licensed")
    with patch.dict(win_license.__salt__, {"cmd.run": mock}):
        win_license.licensed()
        mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /dli")


def test_info():
    """
    Test getting the info about the current license key
    """
    expected = {
        "description": "Prof",
        "licensed": True,
        "name": "Win7",
        "partial_key": "12345",
    }

    mock = MagicMock(
        return_value=(
            "Name: Win7\r\nDescription: Prof\r\nPartial Product Key: 12345\r\n"
            "License Status: Licensed"
        )
    )
    with patch.dict(win_license.__salt__, {"cmd.run": mock}):
        out = win_license.info()
        mock.assert_called_once_with(r"cscript C:\Windows\System32\slmgr.vbs /dli")
        assert out == expected
